from __future__ import annotations

import copy
from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING, overload

from libqtile import configurable
from libqtile.backend.base import Window
from libqtile.command.base import CommandObject, expose_command
from libqtile.command.interface import CommandError
from libqtile.config import ScreenRect

if TYPE_CHECKING:
    from collections.abc import Iterator, Sequence
    from typing import Any, Self

    from libqtile.command.base import ItemT
    from libqtile.group import _Group


class Layout(CommandObject, configurable.Configurable, metaclass=ABCMeta):
    """This class defines the API that should be exposed by all layouts"""

    defaults: list[tuple[str, Any, str]] = []

    def __init__(self, **config: Any) -> None:
        # name is a little odd; we can't resolve it until the class is defined
        # (i.e., we can't figure it out to define it in Layout.defaults), so
        # we resolve it here instead.
        if "name" not in config:
            config["name"] = self.__class__.__name__.lower()

        CommandObject.__init__(self)
        configurable.Configurable.__init__(self, **config)
        self.add_defaults(Layout.defaults)

        self._group: _Group | None = None

    def layout(self, windows: Sequence[Window], screen_rect: ScreenRect) -> None:
        for i in windows:
            self.configure(i, screen_rect)

    def finalize(self) -> None:
        pass

    @property
    def group(self) -> _Group:
        """
        Returns the group this layout is attached to.

        Layouts start out unattached, and are attached when the group is configured and
        each layout is cloned for every group.
        """
        if self._group is None:
            raise RuntimeError("Layout group accessed too early")
        return self._group

    def clone(self, group: _Group) -> Self:
        """Duplicate a layout

        Make a copy of this layout. This is done to provide each group with a
        unique instance of every layout.

        Parameters
        ==========
        group:
            Group to attach new layout instance to.
        """
        c = copy.copy(self)
        c._group = group
        return c

    def _items(self, name: str) -> ItemT:
        if name == "screen" and self.group.screen is not None:
            return True, []
        elif name == "group":
            return True, []
        return None

    def _select(self, name: str, sel: str | int | None) -> CommandObject | None:
        if name == "screen":
            return self.group.screen
        elif name == "group":
            return self.group
        return None

    def show(self, screen_rect: ScreenRect) -> None:
        """Called when layout is being shown"""

    def hide(self) -> None:
        """Called when layout is being hidden"""

    def swap(self, c1: Window, c2: Window) -> None:
        """Swap the two given clients c1 and c2"""
        raise CommandError(f"layout: {self.name} does not support swapping windows")

    def focus(self, client: Window) -> None:
        """Called whenever the focus changes"""

    def blur(self) -> None:
        """Called whenever focus is gone from this layout"""

    @expose_command()
    def info(self) -> dict[str, Any]:
        """Returns a dictionary of layout information"""
        return dict(name=self.name, group=self.group.name if self.group else None)

    @abstractmethod
    def add_client(self, client: Window) -> None:
        """Called whenever a window is added to the group

        Called whether the layout is current or not. The layout should just add
        the window to its internal datastructures, without mapping or
        configuring.
        """

    @abstractmethod
    def remove(self, client: Window) -> Window | None:
        """Called whenever a window is removed from the group

        Called whether the layout is current or not. The layout should just
        de-register the window from its data structures, without unmapping the
        window.

        Returns the "next" window that should gain focus or None.
        """

    @abstractmethod
    def configure(self, client: Window, screen_rect: ScreenRect) -> None:
        """Configure the layout

        This method should:

            - Configure the dimensions and borders of a window using the
              `.place()` method.
            - Call either `.hide()` or `.unhide()` on the window.
        """

    @abstractmethod
    def focus_first(self) -> Window | None:
        """Called when the first client in Layout shall be focused.

        This method should:
            - Return the first client in Layout, if any.
            - Not focus the client itself, this is done by caller.
        """

    @abstractmethod
    def focus_last(self) -> Window | None:
        """Called when the last client in Layout shall be focused.

        This method should:
            - Return the last client in Layout, if any.
            - Not focus the client itself, this is done by caller.
        """

    @abstractmethod
    def focus_next(self, win: Window) -> Window | None:
        """Called when the next client in Layout shall be focused.

        This method should:
            - Return the next client in Layout, if any.
            - Return None if the next client would be the first client.
            - Not focus the client itself, this is done by caller.

        Do not implement a full cycle here, because the Groups cycling relies
        on returning None here if the end of Layout is hit,
        such that Floating clients are included in cycle.

        Parameters
        ==========
        win:
            The currently focused client.
        """

    @abstractmethod
    def focus_previous(self, win: Window) -> Window | None:
        """Called when the previous client in Layout shall be focused.

        This method should:
            - Return the previous client in Layout, if any.
            - Return None if the previous client would be the last client.
            - Not focus the client itself, this is done by caller.

        Do not implement a full cycle here, because the Groups cycling relies
        on returning None here if the end of Layout is hit,
        such that Floating clients are included in cycle.

        Parameters
        ==========
        win:
            The currently focused client.
        """

    @abstractmethod
    def next(self) -> None:
        pass

    @abstractmethod
    def previous(self) -> None:
        pass


class _ClientList:
    """
    ClientList maintains a list of clients and a current client.

    The collection is meant as a base or utility class for special layouts,
    which need to maintain one or several collections of windows, for example
    Columns or Stack, which use this class as base for their internal helper.

    The property 'current_index' get and set the index to the current client,
    whereas 'current_client' property can be used with clients directly.

    The collection implements focus_xxx methods as desired for Group.
    """

    def __init__(self) -> None:
        self._current_idx: int = 0
        self.clients: list[Window] = []

    @property
    def current_index(self) -> int:
        return self._current_idx

    @current_index.setter
    def current_index(self, x: int) -> None:
        if len(self):
            self._current_idx = abs(x % len(self))
        else:
            self._current_idx = 0

    @property
    def current_client(self) -> Window | None:
        if not self.clients:
            return None
        return self.clients[self._current_idx]

    @current_client.setter
    def current_client(self, client: Window) -> None:
        self._current_idx = self.clients.index(client)

    def focus(self, client: Window) -> None:
        """
        Mark the given client as the current focused client in collection.
        This is equivalent to setting current_client.
        """
        self.current_client = client

    def focus_first(self) -> Window | None:
        """
        Returns the first client in collection.
        """
        return self[0]

    def focus_next(self, win: Window) -> Window | None:
        """
        Returns the client next from win in collection.
        """
        try:
            return self[self.index(win) + 1]
        except IndexError:
            return None

    def focus_last(self) -> Window | None:
        """
        Returns the last client in collection.
        """
        return self[-1]

    def focus_previous(self, win: Window) -> Window | None:
        """
        Returns the client previous to win in collection.
        """
        idx = self.index(win)
        if idx > 0:
            return self[idx - 1]
        return None

    def add_client(
        self, client: Window, offset_to_current: int = 0, client_position: str | None = None
    ) -> None:
        """
        Insert the given client into collection at position of the current.

        Use parameter 'offset_to_current' to specify where the client shall be
        inserted. Defaults to zero, which means at position of current client.
        Positive values are after the client.

        Use parameter 'client_position' to insert the given client at 4 specific
        positions: top, bottom, after_current, before_current.
        """
        if client_position is not None:
            if client_position == "after_current":
                return self.add_client(client, offset_to_current=1)
            elif client_position == "before_current":
                return self.add_client(client, offset_to_current=0)
            elif client_position == "top":
                self.append_head(client)
            else:  # ie client_position == "bottom"
                self.append(client)
        else:
            pos = max(0, self._current_idx + offset_to_current)
            if pos < len(self.clients):
                self.clients.insert(pos, client)
            else:
                self.clients.append(client)
        self.current_client = client

    def append_head(self, client: Window) -> None:
        """
        Append the given client in front of list.
        """
        self.clients.insert(0, client)

    def append(self, client: Window) -> None:
        """
        Append the given client to the end of the collection.
        """
        self.clients.append(client)

    def remove(self, client: Window) -> Window | None:
        """
        Remove the given client from collection.
        """
        if client not in self.clients:
            return None
        idx = self.clients.index(client)
        del self.clients[idx]
        if len(self) == 0:
            self._current_idx = 0
        elif idx <= self._current_idx:
            self._current_idx = max(0, self._current_idx - 1)

        return self[self._current_idx]

    def rotate_up(self, maintain_index: bool = True) -> None:
        """
        Rotate the list. The first client is moved to last position.
        If maintain_index is True the current_index is adjusted,
        such that the same client stays current and goes up in list.
        """
        if len(self.clients) > 1:
            self.clients.append(self.clients.pop(0))
            if maintain_index:
                self.current_index -= 1

    def rotate_down(self, maintain_index: bool = True) -> None:
        """
        Rotate the list. The last client is moved to first position.
        If maintain_index is True the current_index is adjusted,
        such that the same client stays current and goes down in list.
        """
        if len(self.clients) > 1:
            self.clients.insert(0, self.clients.pop())
            if maintain_index:
                self.current_index += 1

    def swap(self, c1: Window, c2: Window, focus: int = 1) -> None:
        """
        Swap the two given clients in list.
        The optional argument 'focus' can be 1 or 2.
        In case of 1, the first client c1 is focused, in case of 2 the c2 and
        the current_index is not changed otherwise.
        """
        i1 = self.clients.index(c1)
        i2 = self.clients.index(c2)
        self.clients[i1], self.clients[i2] = self.clients[i2], self.clients[i1]
        if focus == 1:
            self.current_index = i1
        elif focus == 2:
            self.current_index = i2

    def shuffle_up(self, maintain_index: bool = True) -> None:
        """
        Shuffle the list. The current client swaps position with its predecessor.
        If maintain_index is True the current_index is adjusted,
        such that the same client stays current and goes up in list.
        """
        idx = self._current_idx
        if idx > 0:
            self.clients[idx], self.clients[idx - 1] = self.clients[idx - 1], self.clients[idx]
            if maintain_index:
                self.current_index -= 1

    def shuffle_down(self, maintain_index: bool = True) -> None:
        """
        Shuffle the list. The current client swaps position with its successor.
        If maintain_index is True the current_index is adjusted,
        such that the same client stays current and goes down in list.
        """
        idx = self._current_idx
        if idx + 1 < len(self.clients):
            self.clients[idx], self.clients[idx + 1] = self.clients[idx + 1], self.clients[idx]
            if maintain_index:
                self.current_index += 1

    def join(self, other: _ClientList, offset_to_current: int = 0) -> None:
        """
        Add clients from 'other' _ClientList to self.
        'offset_to_current' works as described for add()
        """
        pos = max(0, self.current_index + offset_to_current)
        if pos < len(self.clients):
            self.clients = self.clients[:pos:] + other.clients + self.clients[pos::]
        else:
            self.clients.extend(other.clients)

    def index(self, client: Window) -> int:
        return self.clients.index(client)

    def __len__(self) -> int:
        return len(self.clients)

    @overload
    def __getitem__(self, i: int) -> Window | None: ...

    @overload
    def __getitem__(self, i: slice) -> list[Window]: ...

    def __getitem__(self, i: int | slice) -> Window | None | list[Window]:
        if isinstance(i, slice):
            return self.clients[i]
        try:
            return self.clients[i]
        except IndexError:
            return None

    def __setitem__(self, i: int, value: Window) -> None:
        self.clients[i] = value

    def __iter__(self) -> Iterator[Window]:
        return self.clients.__iter__()

    def __contains__(self, client: Window) -> bool:
        return client in self.clients

    def __str__(self) -> str:
        curr = self.current_client
        return "_ClientList: " + ", ".join(
            [("[%s]" if c == curr else "%s") % c.name for c in self.clients]
        )

    @expose_command()
    def info(self) -> dict[str, Any]:
        return dict(
            clients=[c.name for c in self.clients],
            current=self._current_idx,
        )


class _SimpleLayoutBase(Layout):
    """
    Basic layout class for simple layouts,
    which need to maintain a single list of clients.
    This class offers full fledged list of clients and focus cycling.

    Basic Layouts like Max and Matrix are based on this class
    """

    def __init__(self, **config: Any) -> None:
        Layout.__init__(self, **config)
        self.clients = _ClientList()

    def clone(self, group: _Group) -> Self:
        c = Layout.clone(self, group)
        c.clients = _ClientList()
        return c

    def focus(self, client: Window) -> None:
        self.clients.current_client = client

    def focus_first(self) -> Window | None:
        return self.clients.focus_first()

    def focus_last(self) -> Window | None:
        return self.clients.focus_last()

    def focus_next(self, window: Window) -> Window | None:
        return self.clients.focus_next(window)

    def focus_previous(self, window: Window) -> Window | None:
        return self.clients.focus_previous(window)

    def previous(self) -> None:
        if self.clients.current_client is None:
            return
        client = self.focus_previous(self.clients.current_client) or self.focus_last()
        self.group.focus(client, True)

    def swap(self, window1: Window, window2: Window) -> None:
        self.clients.swap(window1, window2)
        self.group.layout_all()
        self.group.focus(window1)

    def next(self) -> None:
        if self.clients.current_client is None:
            return
        client = self.focus_next(self.clients.current_client) or self.focus_first()
        self.group.focus(client, True)

    def add_client(
        self, client: Window, offset_to_current: int = 0, client_position: str | None = None
    ) -> None:
        self.clients.add_client(client, offset_to_current, client_position)

    def remove(self, client: Window) -> Window | None:
        return self.clients.remove(client)

    def get_windows(self) -> list[Window]:
        return self.clients.clients

    @expose_command()
    def info(self) -> dict[str, Any]:
        d = Layout.info(self)
        d.update(self.clients.info())
        return d



from typing import Any

from libqtile import bar
from libqtile.command.base import expose_command
from libqtile.widget import base


class TextBox(base._TextBox):
    """A flexible textbox that can be updated from bound keys, scripts, and qshell."""

    defaults: list[tuple[str, Any, str]] = [
        ("font", "sans", "Text font"),
        ("fontsize", None, "Font pixel size. Calculated if None."),
        ("fontshadow", None, "font shadow color, default is None(no shadow)"),
        ("padding", None, "Padding left and right. Calculated if None."),
        ("foreground", "#ffffff", "Foreground colour."),
    ]

    def __init__(self, text=" ", width=bar.CALCULATED, **config):
        base._TextBox.__init__(self, text=text, width=width, **config)

    @expose_command()
    def get(self):
        """Retrieve the text in a TextBox widget"""
        return self.text

    @expose_command()
    def update(self, text):
        base._TextBox.update(self, text)



from __future__ import annotations

from typing import TYPE_CHECKING

from libqtile.command.client import InteractiveCommandClient
from libqtile.command.graph import CommandGraphCall, CommandGraphNode
from libqtile.command.interface import CommandInterface
from libqtile.log_utils import logger

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable

    from libqtile.command.graph import SelectorType
    from libqtile.config import _Match


class LazyCall:
    def __init__(self, call: CommandGraphCall, args: tuple, kwargs: dict) -> None:
        """The lazily evaluated command graph call

        Parameters
        ----------
        call: CommandGraphCall
            The call that is made
        args: tuple
            The args passed to the call when it is evaluated.
        kwargs: dict
            The kwargs passed to the call when it is evaluated.
        """
        self._call = call
        self._args = args
        self._kwargs = kwargs

        self._focused: _Match | None = None
        self._if_no_focused: bool = False
        self._layouts: set[str] = set()
        self._when_floating: bool | None = None
        self._condition: bool | None = None
        self._func: Callable[[], bool] = lambda: True

    def __call__(self, *args, **kwargs):
        """Convenience method to allow users to pass arguments to
        functions decorated with `@lazy.function`.

            @lazy.function
            def my_function(qtile, pos_arg, keyword_arg=False):
                pass

            ...

            Key(... my_function("Positional argument", keyword_arg=True))

        """
        # We need to return a new object so the arguments are not shared between
        # a single instance of the LazyCall object.
        return LazyCall(self._call, (*self._args, *args), {**self._kwargs, **kwargs})

    @property
    def selectors(self) -> list[SelectorType]:
        """The selectors for the given call"""
        return self._call.selectors

    @property
    def name(self) -> str:
        """The name of the given call"""
        return self._call.name

    @property
    def args(self) -> tuple:
        """The args to the given call"""
        return self._args

    @property
    def kwargs(self) -> dict:
        """The kwargs to the given call"""
        return self._kwargs

    def when(
        self,
        focused: _Match | None = None,
        if_no_focused: bool = False,
        layout: Iterable[str] | str | None = None,
        when_floating: bool | None = None,
        func: Callable | None = None,
        condition: bool | None = None,
    ) -> LazyCall:
        """Enable call only for matching criteria.

        Keyword parameters
        ----------
        focused: Match or None
            Match criteria to enable call for the current window.
        if_no_focused: bool
            Whether or not the `focused` attribute should also
            match when there is no focused window.
            This is useful when the `focused` attribute is e.g. set
            to a regex that should also match when there is
            no focused window.
            By default this is set to `False` so that the focused
            attribute only matches when there is actually a focused window.
        layout: str, Iterable[str], or None
            Restrict call to one or more layouts.
            If None, enable the call for all layouts.
        when_floating: bool
            Enable call when the current window is floating.
        func: callable
            Enable call when the result of the callable evaluates to True
        condition: a boolean value to determine whether the lazy object should
            be run. Unlike 'func', the condition is evaluated once when the config
            file is first loaded.

        """
        self._focused = focused

        self._if_no_focused = if_no_focused

        self._condition = condition

        if func is not None:
            self._func = func

        if layout is not None:
            self._layouts = {layout} if isinstance(layout, str) else set(layout)

        self._when_floating = when_floating
        return self

    def check(self, q) -> bool:
        cur_win_floating = q.current_window and q.current_window.floating

        if self._condition is False:
            return False

        if self._focused:
            if q.current_window and not self._focused.compare(q.current_window):
                return False

            if not q.current_window and not self._if_no_focused:
                return False

        if cur_win_floating and self._when_floating is False:
            return False

        if not cur_win_floating and self._when_floating:
            return False

        if self._layouts and q.current_layout.name not in self._layouts:
            return False

        if self._func is not None:
            try:
                result = self._func()
            except Exception:
                logger.exception("Error when running function in lazy call. Ignoring.")
                result = True

            if not result:
                return False

        return True


class LazyCommandInterface(CommandInterface):
    """A lazy loading command object

    Allows all commands and items to be resolved at run time, and returns
    lazily evaluated commands.
    """

    def execute(self, call: CommandGraphCall, args: tuple, kwargs: dict) -> LazyCall:
        """Lazily evaluate the given call"""
        return LazyCall(call, args, kwargs)

    def has_command(self, node: CommandGraphNode, command: str) -> bool:
        """Lazily resolve the given command"""
        return True

    def has_item(self, node: CommandGraphNode, object_type: str, item: str | int) -> bool:
        """Lazily resolve the given item"""
        return True


lazy = InteractiveCommandClient(LazyCommandInterface())



from typing import Any

try:
    from dbus_fast.aio import MessageBus
    from dbus_fast.constants import NameFlag, RequestNameReply
    from dbus_fast.service import ServiceInterface, method, signal

    has_dbus = True
except ImportError:
    has_dbus = False

from libqtile.log_utils import logger

BUS_NAME = "org.freedesktop.Notifications"
SERVICE_PATH = "/org/freedesktop/Notifications"

notifier: Any = None


class ClosedReason:
    expired = 1
    dismissed = 2
    method = 3  # CloseNotification method


if has_dbus:

    class NotificationService(ServiceInterface):
        def __init__(self, manager):
            super().__init__(BUS_NAME)
            self.manager = manager
            self._capabilities = {"body"}

        @method()
        def GetCapabilities(self) -> "as":  # type:ignore  # noqa: N802, F722
            return list(self._capabilities)

        def register_capabilities(self, capabilities):
            if isinstance(capabilities, str):
                self._capabilities.add(capabilities)
            elif isinstance(capabilities, tuple | list | set):
                self._capabilities.update(set(capabilities))

        @method()
        def Notify(  # noqa: N802, F722
            self,
            app_name: "s",  # type:ignore  # noqa: F821
            replaces_id: "u",  # type:ignore  # noqa: F821
            app_icon: "s",  # type:ignore  # noqa: F821
            summary: "s",  # type:ignore  # noqa: F821
            body: "s",  # type:ignore  # noqa: F821
            actions: "as",  # type:ignore  # noqa: F722, F821
            hints: "a{sv}",  # type:ignore  # noqa: F722, F821
            timeout: "i",  # type:ignore  # noqa: F821
        ) -> "u":  # type:ignore  # noqa: F821
            notif = Notification(
                summary, body, timeout, hints, app_name, replaces_id, app_icon, actions
            )
            return self.manager.add(notif)

        @method()
        def CloseNotification(self, nid: "u"):  # type:ignore  # noqa: F821, N802
            self.manager.close(nid)

        @signal()
        def NotificationClosed(  # noqa: N802
            self,
            nid: "u",  # type: ignore  # noqa: F821
            reason: "u",  # type: ignore  # noqa: F821
        ) -> "uu":  # type: ignore  # noqa: F821
            return [nid, reason]

        @signal()
        def ActionInvoked(  # noqa: N802
            self,
            nid: "u",  # type: ignore  # noqa: F821
            action_key: "s",  # type: ignore  # noqa: F821
        ) -> "us":  # type:ignore  # noqa: N802, F821
            return [nid, action_key]

        @method()
        def GetServerInformation(self) -> "ssss":  # type: ignore  # noqa: N802, F821
            return ["qtile-notify-daemon", "qtile", "1.0", "1"]

    class Notification:
        def __init__(
            self,
            summary,
            body="",
            timeout=-1,
            hints=None,
            app_name="",
            replaces_id=None,
            app_icon=None,
            actions=None,
        ):
            self.summary = summary
            self.body = body
            self.timeout = timeout
            self.hints = hints or {}
            self.app_name = app_name
            self.replaces_id = replaces_id
            self.app_icon = app_icon
            self.actions = actions

    class NotificationManager:
        def __init__(self):
            self.notifications = []
            self.callbacks = []
            self.close_callbacks = []
            self._service = None
            self.bus = None

        async def service(self):
            if not self.callbacks:
                if self.bus is None:
                    try:
                        self.bus = await MessageBus().connect()
                    except Exception:
                        logger.exception("Dbus connection failed")
                        self._service = None
                        return self._service

                    self._service = NotificationService(self)
                    self.bus.export(SERVICE_PATH, self._service)

                reply = await self.bus.request_name(
                    BUS_NAME,
                    flags=NameFlag.ALLOW_REPLACEMENT
                    | NameFlag.REPLACE_EXISTING
                    | NameFlag.DO_NOT_QUEUE,
                )

                # Check the reply to see if another server is already running.
                # Other replies could be RequestNameReply.PRIMARY_OWNER or
                # RequestNameReply.ALREADY_ONWER. Both would indicate qtile server
                # is running successfully so we don't need to check for them.
                # RequestNameReply.IN_QUEUE should not be possible as we pass
                # NameFlag.DO_NOT_QUEUE when requesting the name.
                if reply == RequestNameReply.EXISTS:
                    logger.warning(
                        "Cannot start notification server as another server is already running."
                    )
                    self._service = None
                    self.bus.disconnect()
                    self.bus = None

            return self._service

        async def register(self, callback, capabilities=None, on_close=None):
            service = await self.service()
            if not service:
                logger.warning(
                    "Registering %s without any dbus connection existing",
                    callback.__name__,
                )
            self.callbacks.append(callback)
            if capabilities:
                self._service.register_capabilities(capabilities)
            if on_close:
                self.close_callbacks.append(on_close)

        def unregister(self, callback, on_close=None):
            try:
                self.callbacks.remove(callback)
            except ValueError:
                logger.error("Unable to remove notify callback. Unknown callback.")

            if on_close:
                try:
                    self.close_callbacks.remove(on_close)
                except ValueError:
                    logger.error("Unable to remove notify on_close callback. Unknown callback.")

        def add(self, notif):
            self.notifications.append(notif)
            notif.id = len(self.notifications)
            for callback in self.callbacks:
                try:
                    callback(notif)
                except Exception:
                    logger.exception("Exception in notifier callback")
            return len(self.notifications)

        def show(self, *args, **kwargs):
            notif = Notification(*args, **kwargs)
            return (notif, self.add(notif))

        def close(self, nid):
            for callback in self.close_callbacks:
                try:
                    callback(nid)
                except Exception:
                    logger.exception("Exception in notifier close callback")

    notifier = NotificationManager()



from __future__ import annotations

import asyncio
import re
import string
from typing import TYPE_CHECKING

from dbus_fast import Message, Variant
from dbus_fast.aio import MessageBus
from dbus_fast.constants import MessageType

from libqtile import pangocffi
from libqtile.command.base import expose_command
from libqtile.log_utils import logger
from libqtile.utils import _send_dbus_message, add_signal_receiver, create_task
from libqtile.widget import base

if TYPE_CHECKING:
    from typing import Any

MPRIS_PATH = "/org/mpris/MediaPlayer2"
MPRIS_OBJECT = "org.mpris.MediaPlayer2"
MPRIS_PLAYER = "org.mpris.MediaPlayer2.Player"
PROPERTIES_INTERFACE = "org.freedesktop.DBus.Properties"
MPRIS_REGEX = re.compile(r"(\{(.*?):(.*?)(:.*?)?\})")


class Mpris2Formatter(string.Formatter):
    """
    Custom string formatter for MPRIS2 metadata.

    Keys have a colon (e.g. "xesam:title") which causes issues with python's string
    formatting as the colon splits the identifier from the format specification.

    This formatter handles this issue by changing the first colon to an underscore and
    then formatting the incoming kwargs to match.

    Additionally, a default value is returned when an identifier is not provided by the
    kwarg data.
    """

    def __init__(self, default=""):
        string.Formatter.__init__(self)
        self._default = default

    def get_value(self, key, args, kwargs):
        """
        Replaces colon in kwarg keys with an underscore before getting value.

        Missing identifiers are replaced with the default value.
        """
        kwargs = {k.replace(":", "_"): v for k, v in kwargs.items()}
        try:
            return pangocffi.markup_escape_text(
                string.Formatter.get_value(self, key, args, kwargs)
            )
        except (IndexError, KeyError):
            return self._default

    def parse(self, format_string):
        """
        Replaces first colon in format string with an underscore.

        This will cause issues if any identifier is provided that does not
        contain a colon. This should not happen according to the MPRIS2
        specification!
        """
        format_string = MPRIS_REGEX.sub(r"{\2_\3\4}", format_string)
        return string.Formatter.parse(self, format_string)


class Mpris2(base._TextBox):
    """An MPRIS 2 widget

    A widget which displays the current track/artist of your favorite MPRIS
    player. This widget scrolls the text if neccessary and information that
    is displayed is configurable.

    The widget relies on players broadcasting signals when the metadata or playback
    status changes. If you are getting inconsistent results then you can enable background
    polling of the player by setting the `poll_interval` parameter. This is disabled by
    default.

    Basic mouse controls are also available: button 1 = play/pause,
    scroll up = next track, scroll down = previous track.

    Widget requirements: dbus-fast_.

    .. _dbus-fast: https://pypi.org/project/dbus-fast/
    """

    defaults = [
        ("name", "audacious", "Name of the MPRIS widget."),
        (
            "objname",
            None,
            "DBUS MPRIS 2 compatible player identifier"
            "- Find it out with dbus-monitor - "
            "Also see: http://specifications.freedesktop.org/"
            "mpris-spec/latest/#Bus-Name-Policy. "
            "``None`` will listen for notifications from all MPRIS2 compatible players.",
        ),
        (
            "format",
            "{xesam:title} - {xesam:album} - {xesam:artist}",
            "Format string for displaying metadata. "
            "See http://www.freedesktop.org/wiki/Specifications/mpris-spec/metadata/#index5h3 "
            "for available values. The special filed '{qtile:player}' can be used to display the "
            "player name.",
        ),
        ("separator", ", ", "Separator for metadata fields that are a list."),
        (
            "display_metadata",
            ["xesam:title", "xesam:album", "xesam:artist"],
            "(Deprecated) Which metadata identifiers to display. ",
        ),
        ("scroll", True, "Whether text should scroll."),
        ("playing_text", "{track}", "Text to show when playing"),
        ("paused_text", "Paused: {track}", "Text to show when paused"),
        ("stopped_text", "", "Text to show when stopped"),
        (
            "stop_pause_text",
            None,
            "(Deprecated) Optional text to display when in the stopped/paused state",
        ),
        (
            "no_metadata_text",
            "No metadata for current track",
            "Text to show when track has no metadata",
        ),
        (
            "poll_interval",
            0,
            "Periodic background polling interval of player (0 to disable polling).",
        ),
    ]

    def __init__(self, **config):
        base._TextBox.__init__(self, "", **config)
        self.add_defaults(Mpris2.defaults)
        self.is_playing = False
        self.count = 0
        self.displaytext = ""
        self.track_info = ""
        self.status = "{track}"
        self.add_callbacks(
            {
                "Button1": self.play_pause,
                "Button4": self.next,
                "Button5": self.previous,
            }
        )
        paused = ""
        stopped = ""
        if "stop_pause_text" in config:
            logger.warning(
                "The use of 'stop_pause_text' is deprecated. Please use 'paused_text' and 'stopped_text' instead."
            )
            if "paused_text" not in config:
                paused = self.stop_pause_text

            if "stopped_text" not in config:
                stopped = self.stop_pause_text

        if "display_metadata" in config:
            logger.warning(
                "The use of `display_metadata is deprecated. Please use `format` instead."
            )
            self.format = " - ".join(f"{{{s}}}" for s in config["display_metadata"])

        self._formatter = Mpris2Formatter()

        self.prefixes = {
            "Playing": self.playing_text,
            "Paused": paused or self.paused_text,
            "Stopped": stopped or self.stopped_text,
        }

        self._current_player: str | None = None
        self.player_names: dict[str, str] = {}
        self._background_poll: asyncio.TimerHandle | None = None
        self.bus: MessageBus | None = None

    @property
    def player(self) -> str:
        if self._current_player is None:
            return "None"
        else:
            return self.player_names.get(self._current_player, "Unknown")

    async def _config_async(self):
        # These two listeners create separate bus connections. Each connection only has one
        # callback so we don't need any logic to identify the message and the appropriate
        # handler in this code.

        # Set up a listener for NameOwner changes so we can remove players when they close
        await add_signal_receiver(
            self._name_owner_changed,
            session_bus=True,
            signal_name="NameOwnerChanged",
            dbus_interface="org.freedesktop.DBus",
        )

        # Listen out for signals from any Mpris2 compatible player
        subscribe = await add_signal_receiver(
            self.message,
            session_bus=True,
            signal_name="PropertiesChanged",
            bus_name=self.objname,
            path="/org/mpris/MediaPlayer2",
            dbus_interface="org.freedesktop.DBus.Properties",
        )

        if not subscribe:
            logger.warning("Unable to add signal receiver for Mpris2 players")

        # If the user has specified a player to be monitored, we can poll it now.
        if self.objname is not None:
            await self._check_player()

    def _name_owner_changed(self, message):
        # We need to track when an interface has been removed from the bus
        # We use the NameOwnerChanged signal and check if the new owner is
        # empty.
        name, _, new_owner = message.body

        # Check if the current player has closed
        if new_owner == "" and name == self._current_player:
            self._current_player = None
            self.update("")

            # Cancel any scheduled background poll
            self._set_background_poll(False)

    def message(self, message):
        create_task(self.process_message(message))

    async def process_message(self, message):
        current_player = message.sender

        if current_player not in self.player_names:
            self.player_names[current_player] = await self.get_player_name(current_player)

        self._current_player = current_player

        self.parse_message(*message.body)

    async def _send_message(self, destination, interface, path, member, signature, body):
        bus, message = await _send_dbus_message(
            session_bus=True,
            message_type=MessageType.METHOD_CALL,
            destination=destination,
            interface=interface,
            path=path,
            member=member,
            signature=signature,
            body=body,
            bus=self.bus,
        )

        # We should reuse the same bus connection for repeated calls.
        if self.bus is None:
            self.bus = bus

        return message

    async def _check_player(self):
        """Check for player at startup and retrieve metadata."""
        if not (self.objname or self._current_player):
            return

        message = await self._send_message(
            self.objname if self.objname else self._current_player,
            PROPERTIES_INTERFACE,
            MPRIS_PATH,
            "GetAll",
            "s",
            [MPRIS_PLAYER],
        )

        # If we get an error here it will be because the player object doesn't exist
        if message.message_type != MessageType.METHOD_RETURN:
            self._current_player = None
            self.update("")
            return

        if message.body:
            self._current_player = message.sender
            self.parse_message(self.objname, message.body[0], [])

    def _set_background_poll(self, poll=True):
        if self._background_poll is not None:
            self._background_poll.cancel()

        if poll:
            self._background_poll = self.timeout_add(self.poll_interval, self._check_player)

    async def get_player_name(self, player):
        message = await self._send_message(
            player,
            PROPERTIES_INTERFACE,
            MPRIS_PATH,
            "Get",
            "ss",
            [MPRIS_OBJECT, "Identity"],
        )

        if message.message_type != MessageType.METHOD_RETURN:
            logger.warning("Could not retrieve identity of player on %s.", player)
            return ""

        return message.body[0].value

    def parse_message(
        self,
        _interface_name: str,
        changed_properties: dict[str, Any],
        _invalidated_properties: list[str],
    ) -> None:
        """
        http://specifications.freedesktop.org/mpris-spec/latest/Track_List_Interface.html#Mapping:Metadata_Map
        """
        if not self.configured:
            return

        if "Metadata" not in changed_properties and "PlaybackStatus" not in changed_properties:
            return

        self.displaytext = ""

        metadata = changed_properties.get("Metadata")
        if metadata:
            self.track_info = self.get_track_info(metadata.value)

        playbackstatus = getattr(changed_properties.get("PlaybackStatus"), "value", None)
        if playbackstatus:
            self.is_playing = playbackstatus == "Playing"
            self.status = self.prefixes.get(playbackstatus, "{track}")

        if not self.track_info:
            self.track_info = self.no_metadata_text

        self.displaytext = self.status.format(track=self.track_info)

        if self.text != self.displaytext:
            self.update(self.displaytext)

        if self.poll_interval:
            self._set_background_poll()

    def get_track_info(self, metadata: dict[str, Variant]) -> str:
        self.metadata = {}
        for key in metadata:
            new_key = key
            val = getattr(metadata.get(key), "value", None)
            if isinstance(val, str):
                self.metadata[new_key] = val
            elif isinstance(val, list):
                self.metadata[new_key] = self.separator.join(y for y in val if isinstance(y, str))
        if self.player is not None:
            self.metadata["qtile:player"] = self.player

        return self._formatter.format(self.format, **self.metadata).replace("\n", "")

    def _player_cmd(self, cmd: str) -> None:
        if self._current_player is None:
            return

        task = create_task(self._send_player_cmd(cmd))
        assert task
        task.add_done_callback(self._task_callback)

    async def _send_player_cmd(self, cmd: str) -> Message | None:
        message = await self._send_message(
            self._current_player,
            MPRIS_PLAYER,
            MPRIS_PATH,
            cmd,
            "",
            [],
        )

        return message

    def _task_callback(self, task: asyncio.Task) -> None:
        message = task.result()

        # This happens if we can't connect to dbus. Logger call is made
        # elsewhere so we don't need to do any more here.
        if message is None:
            return

        if message.message_type != MessageType.METHOD_RETURN:
            logger.warning("Unable to send command to player.")

    @expose_command()
    def play_pause(self) -> None:
        """Toggle the playback status."""
        self._player_cmd("PlayPause")

    @expose_command()
    def next(self) -> None:
        """Play the next track."""
        self._player_cmd("Next")

    @expose_command()
    def previous(self) -> None:
        """Play the previous track."""
        self._player_cmd("Previous")

    @expose_command()
    def stop(self) -> None:
        """Stop playback."""
        self._player_cmd("Stop")

    @expose_command()
    def info(self):
        """What's the current state of the widget?"""
        d = base._TextBox.info(self)
        d.update(dict(isplaying=self.is_playing, player=self.player))
        return d



from __future__ import annotations

from typing import TYPE_CHECKING

from libqtile.command.base import expose_command
from libqtile.layout.base import _SimpleLayoutBase

if TYPE_CHECKING:
    from typing import Self

    from libqtile.backend.base import Window
    from libqtile.group import _Group


class VerticalTile(_SimpleLayoutBase):
    """Tiling layout that works nice on vertically mounted monitors

    The available height gets divided by the number of panes, if no pane is
    maximized. If one pane has been maximized, the available height gets split
    in master and secondary area. The maximized pane (master pane) gets the
    full height of the master area and the other panes (secondary panes) share
    the remaining space.  The master area (at default 75%) can grow and shrink
    via keybindings.

    ::

        -----------------                -----------------  ---
        |               |                |               |   |
        |       1       |  <-- Panes     |               |   |
        |               |        |       |               |   |
        |---------------|        |       |               |   |
        |               |        |       |               |   |
        |       2       |  <-----+       |       1       |   |  Master Area
        |               |        |       |               |   |
        |---------------|        |       |               |   |
        |               |        |       |               |   |
        |       3       |  <-----+       |               |   |
        |               |        |       |               |   |
        |---------------|        |       |---------------|  ---
        |               |        |       |       2       |   |
        |       4       |  <-----+       |---------------|   |  Secondary Area
        |               |                |       3       |   |
        -----------------                -----------------  ---

        Normal behavior.                 One maximized pane in the master area
        No maximized pane.               and two secondary panes in the
        No specific areas.               secondary area.

    ::

        -----------------------------------  In some cases, VerticalTile can be
        |                                 |  useful on horizontal mounted
        |                1                |  monitors too.
        |                                 |  For example, if you want to have a
        |---------------------------------|  web browser and a shell below it.
        |                                 |
        |                2                |
        |                                 |
        -----------------------------------


    Suggested keybindings:

    ::

        Key([modkey], 'j', lazy.layout.down()),
        Key([modkey], 'k', lazy.layout.up()),
        Key([modkey], 'Tab', lazy.layout.next()),
        Key([modkey, 'shift'], 'Tab', lazy.layout.next()),
        Key([modkey, 'shift'], 'j', lazy.layout.shuffle_down()),
        Key([modkey, 'shift'], 'k', lazy.layout.shuffle_up()),
        Key([modkey], 'm', lazy.layout.maximize()),
        Key([modkey], 'n', lazy.layout.normalize()),
    """

    defaults = [
        ("border_focus", "#FF0000", "Border color(s) for the focused window."),
        ("border_normal", "#FFFFFF", "Border color(s) for un-focused windows."),
        ("border_width", 1, "Border width."),
        ("single_border_width", None, "Border width for single window."),
        ("single_margin", None, "Margin for single window."),
        ("margin", 0, "Border margin (int or list of ints [N E S W])."),
    ]

    ratio = 0.75
    steps = 0.05

    def __init__(self, **config):
        _SimpleLayoutBase.__init__(self, **config)
        self.add_defaults(VerticalTile.defaults)
        if self.single_border_width is None:
            self.single_border_width = self.border_width
        if self.single_margin is None:
            self.single_margin = self.margin
        self.maximized = None

    def add_client(self, window):
        return self.clients.add_client(window, 1)

    def remove(self, window: Window) -> Window | None:
        if self.maximized is window:
            self.maximized = None
        return self.clients.remove(window)

    def clone(self, group: _Group) -> Self:
        c = _SimpleLayoutBase.clone(self, group)
        c.maximized = None
        return c

    def configure(self, window, screen_rect):
        if self.clients and window in self.clients:
            n = len(self.clients)
            index = self.clients.index(window)

            # border
            border_width = self.border_width if n > 1 else self.single_border_width
            border_color = self.border_focus if window.has_focus else self.border_normal

            # margin
            margin = self.margin if n > 1 else self.single_margin

            # width
            width = screen_rect.width - border_width * 2

            # y
            y = screen_rect.y

            # height
            if n > 1:
                main_area_height = int(screen_rect.height * self.ratio)
                sec_area_height = screen_rect.height - main_area_height

                main_pane_height = main_area_height - border_width * 2
                sec_pane_height = sec_area_height // (n - 1) - border_width * 2
                normal_pane_height = (screen_rect.height // n) - (border_width * 2)

                if self.maximized:
                    y += (index * sec_pane_height) + (border_width * 2 * index)
                    if window is self.maximized:
                        height = main_pane_height
                    else:
                        height = sec_pane_height
                        if index > self.clients.index(self.maximized):
                            y = y - sec_pane_height + main_pane_height
                else:
                    height = normal_pane_height
                    y += (index * normal_pane_height) + (border_width * 2 * index)
            else:
                height = screen_rect.height - 2 * border_width

            window.place(
                screen_rect.x, y, width, height, border_width, border_color, margin=margin
            )
            window.unhide()
        else:
            window.hide()

    def _grow(self):
        if self.ratio + self.steps < 1:
            self.ratio += self.steps
            self.group.layout_all()

    def _shrink(self):
        if self.ratio - self.steps > 0:
            self.ratio -= self.steps
            self.group.layout_all()

    @expose_command("up")
    def previous(self) -> None:
        _SimpleLayoutBase.previous(self)

    @expose_command("down")
    def next(self) -> None:
        _SimpleLayoutBase.next(self)

    @expose_command()
    def shuffle_up(self):
        self.clients.shuffle_up()
        self.group.layout_all()

    @expose_command()
    def shuffle_down(self):
        self.clients.shuffle_down()
        self.group.layout_all()

    @expose_command()
    def maximize(self):
        if self.clients:
            self.maximized = self.clients.current_client
            self.group.layout_all()

    @expose_command()
    def normalize(self):
        self.maximized = None
        self.group.layout_all()

    @expose_command()
    def grow(self):
        if not self.maximized:
            return
        if self.clients.current_client is self.maximized:
            self._grow()
        else:
            self._shrink()

    @expose_command()
    def shrink(self):
        if not self.maximized:
            return
        if self.clients.current_client is self.maximized:
            self._shrink()
        else:
            self._grow()



from __future__ import annotations

import argparse
import os
import os.path
import shutil
import sys
from glob import glob
from pathlib import Path
from typing import TYPE_CHECKING

from libqtile.scripts.migrations import MIGRATIONS, load_migrations
from libqtile.utils import get_config_file

if TYPE_CHECKING:
    from collections.abc import Iterator

    import libcst

BACKUP_SUFFIX = ".migrate.bak"


class AbortMigration(Exception):
    pass


class SkipFile(Exception):
    pass


def needs_libcst(func):
    def _wrapper(*args, **kwargs):
        if "libcst" not in sys.modules:
            try:
                global libcst
                libcst = __import__("libcst", globals(), locals())
            except ImportError:
                print("libcst is needed for 'qtile migrate' commands.")
                print("Please install it and try again.")
                sys.exit(1)
        func(*args, **kwargs)

    return _wrapper


def version_tuple(value: str) -> tuple[int, ...]:
    try:
        val = tuple(int(x) for x in value.split("."))
        return val
    except TypeError:
        raise argparse.ArgumentTypeError("Cannot parse version string.")


def file_and_backup(config_dir: str) -> Iterator[tuple[str, str]]:
    if os.path.isfile(config_dir):
        config_dir = os.path.dirname(config_dir)
    for py in glob(os.path.join(config_dir, "*.py")):
        backup = py + BACKUP_SUFFIX
        yield py, backup


class QtileMigrate:
    """
    May be overkill to use a class here but we can store state (i.e. args)
    without needing to pass them around all the time.
    """

    @needs_libcst
    def __call__(self, args: argparse.Namespace) -> None:
        """
        This is called by ArgParse when we run `qtile migrate`. The parsed options are
        passed as an argument.
        """
        self.args = args
        self.filter_migrations()

        if self.args.list_migrations:
            self.list_migrations()
            return
        elif self.args.info:
            self.show_migration_info()
            return
        else:
            self.run_migrations()

    def filter_migrations(self) -> None:
        load_migrations()
        if self.args.run_migrations:
            self.migrations = [m for m in MIGRATIONS if m.ID in self.args.run_migrations]
        elif self.args.after_version:
            self.migrations = [m for m in MIGRATIONS if m.get_version() > self.args.after_version]
        else:
            self.migrations = MIGRATIONS

        if not self.migrations:
            sys.exit("No migrations found.")

    def list_migrations(self) -> None:
        width = max(len(m.ID) for m in self.migrations) + 4

        ordered = sorted(self.migrations, key=lambda m: (m.get_version(), m.ID))

        print(f"ID{' ' * (width - 2)}{'After Version':<15}Summary")
        for m in ordered:
            summary = m.show_summary().replace("``", "'")
            print(f"{m.ID:<{width}}{m.AFTER_VERSION:^15}{summary}")

    def show_migration_info(self) -> None:
        migration_id = self.args.info
        migration = [m for m in MIGRATIONS if m.ID == migration_id]
        if not migration:
            print(f"Unknown migration: {migration_id}")
            sys.exit(1)

        print(f"{migration_id}:")
        print(migration[0].show_help())

    def get_source(self, path: str) -> libcst.metadata.MetadataWrapper:
        module = libcst.parse_module(Path(path).read_text())
        return libcst.metadata.MetadataWrapper(module)

    def lint(self, path: str) -> None:
        print(f"{path}:")
        source = self.get_source(path)
        lint_lines = []

        for m in self.migrations:
            migrator = m()
            migrator.migrate(source)
            lint_lines.extend(migrator.show_lint())

        lint_lines.sort()
        print("\n".join(map(str, lint_lines)))

    def migrate(self, path: str) -> bool:
        source: libcst.metadata.MetadataWrapper | libcst.Module = self.get_source(path)
        changed = False

        for m in self.migrations:
            migrator = m()
            migrator.migrate(source)

            diff = migrator.show_diff(self.args.no_colour)

            if diff:
                if self.args.show_diff or not self.args.yes:
                    print(f"{m.ID}: {m.show_summary()}\n")
                    print(f"{diff}\n")

                if self.args.yes:
                    assert migrator.updated is not None
                    source = libcst.metadata.MetadataWrapper(migrator.updated)
                    changed = True

                else:
                    while (
                        a := input("Apply changes? (y)es, (n)o, (s)kip file, (q)uit. ").lower()
                    ) not in ("y", "n", "s", "q"):
                        print("Unexpected response. Try again.")

                    if a == "y":
                        assert migrator.updated is not None
                        source = libcst.metadata.MetadataWrapper(migrator.updated)
                        changed = True
                    elif a == "n":
                        assert migrator.original is not None
                        source = migrator.original
                    elif a == "s":
                        raise SkipFile
                    elif a == "q":
                        raise AbortMigration

        if not changed:
            return False

        if not self.args.yes:
            while (save := input(f"Save all changes to {path}? (y)es, (n)o. ").lower()) not in (
                "y",
                "n",
            ):
                print("Unexpected response. Try again.")
            do_save = save == "y"
        else:
            do_save = True

        if do_save:
            if isinstance(source, libcst.metadata.MetadataWrapper):
                source = source.module
            Path(f"{path}").write_text(source.code)
            print("Saved!")
            return True
        else:
            return False

    def run_migrations(self) -> None:
        backups = []
        changed_files = []
        aborted = False

        for py, backup in file_and_backup(self.args.config):
            if self.args.lint:
                self.lint(py)
                continue
            else:
                try:
                    shutil.copyfile(py, backup)
                    backups.append(backup)
                    changed = self.migrate(py)
                    if changed:
                        changed_files.append(py)
                except SkipFile:
                    backups.remove(backup)
                    continue
                except AbortMigration:
                    aborted = True
                    break

        if aborted:
            print("Migration aborted. Reverting changes.")
            for f in changed_files:
                shutil.copyfile(f + BACKUP_SUFFIX, f)

        elif backups:
            print("Finished. Backup files have not been deleted.")


def add_subcommand(subparsers, parents):
    parser = subparsers.add_parser(
        "migrate", parents=parents, help="Migrate a configuration file to the current API."
    )
    parser.add_argument(
        "-c",
        "--config",
        action="store",
        default=get_config_file(),
        help="Use the specified configuration file (migrates every .py file in this directory).",
    )
    parser.add_argument(
        "--yes",
        action="store_true",
        help="Automatically apply diffs with no confirmation.",
    )
    parser.add_argument(
        "--show-diff", action="store_true", help="When used with --yes, will still output diff."
    )
    parser.add_argument(
        "--lint", action="store_true", help="Providing linting output but don't update config."
    )
    parser.add_argument(
        "--list-migrations", action="store_true", help="List available migrations."
    )
    parser.add_argument(
        "--info",
        metavar="ID",
        help="Show detailed info for the migration with the given ID",
    )
    parser.add_argument(
        "--after-version",
        metavar="VERSION",
        type=version_tuple,
        help="Run migrations introduced after VERSION.",
    )
    parser.add_argument(
        "-r",
        "--run-migrations",
        type=lambda value: value.split(","),
        metavar="ID",
        help="Run named migration[s]. Comma separated list for multiple migrations",
    )
    parser.add_argument(
        "--no-colour", action="store_true", help="Do not use colour in diff output."
    )
    parser.add_argument("-v", "--verbose", action="store_true", help="Increase output verbosity")
    parser.set_defaults(func=QtileMigrate())



import re

from libqtile.log_utils import logger
from libqtile.utils import add_signal_receiver
from libqtile.widget import base


class KeyboardKbdd(base.BackgroundPoll):
    """Widget for changing keyboard layouts per window, using kbdd

    kbdd should be installed and running, you can get it from:
    https://github.com/qnikst/kbdd

    The widget also requires dbus-fast_.

    .. _dbus-fast: https://pypi.org/project/dbus-fast/
    """

    defaults = [
        ("update_interval", 1, "Update interval in seconds."),
        (
            "configured_keyboards",
            ["us", "ir"],
            "your predefined list of keyboard layouts.example: ['us', 'ir', 'es']",
        ),
        (
            "colours",
            None,
            "foreground colour for each layout"
            "either 'None' or a list of colours."
            "example: ['ffffff', 'E6F0AF']. ",
        ),
    ]

    def __init__(self, **config):
        base.BackgroundPoll.__init__(self, "", **config)
        self.add_defaults(KeyboardKbdd.defaults)
        self.keyboard = self.configured_keyboards[0]
        self.is_kbdd_running = self._check_kbdd()
        if not self.is_kbdd_running:
            self.keyboard = "N/A"

    def _check_kbdd(self):
        try:
            running_list = self.call_process(["ps", "axw"])
        except FileNotFoundError:
            logger.error("'ps' is not installed. Cannot check if kbdd is running.")
            return False

        if re.search("kbdd", running_list):
            self.keyboard = self.configured_keyboards[0]
            return True

        logger.error("kbdd is not running.")
        return False

    async def _config_async(self):
        subscribed = await add_signal_receiver(
            self._signal_received,
            session_bus=True,
            signal_name="layoutChanged",
            dbus_interface="ru.gentoo.kbdd",
        )

        if not subscribed:
            logger.warning("Could not subscribe to kbdd signal.")

    def _signal_received(self, message):
        self._layout_changed(*message.body)

    def _layout_changed(self, layout_changed):
        """
        Handler for "layoutChanged" dbus signal.
        """
        if self.colours:
            self._set_colour(layout_changed)
        self.keyboard = self.configured_keyboards[layout_changed]

    def _set_colour(self, index):
        if isinstance(self.colours, list):
            try:
                self.layout.colour = self.colours[index]
            except IndexError:
                self._set_colour(index - 1)
        else:
            logger.error(
                'variable "colours" should be a list, to set a\
                            colour for all layouts, use "foreground".'
            )

    def poll(self):
        if not self.is_kbdd_running:
            if self._check_kbdd():
                self.is_kbdd_running = True
                return self.configured_keyboards[0]
        return self.keyboard



from libqtile import bar, hook
from libqtile.widget import base


class CurrentScreen(base._TextBox):
    """Indicates whether the screen this widget is on is currently active or not"""

    defaults = [
        ("active_text", "A", "Text displayed when the screen is active"),
        ("inactive_text", "I", "Text displayed when the screen is inactive"),
        ("active_color", "00ff00", "Color when screen is active"),
        ("inactive_color", "ff0000", "Color when screen is inactive"),
    ]

    def __init__(self, width=bar.CALCULATED, **config):
        base._TextBox.__init__(self, "", width, **config)
        self.add_defaults(CurrentScreen.defaults)

    def _configure(self, qtile, bar):
        base._TextBox._configure(self, qtile, bar)
        hook.subscribe.current_screen_change(self.update_text)
        self.update_text()

    def update_text(self):
        if self.qtile.current_screen == self.bar.screen:
            self.layout.colour = self.active_color
            self.update(self.active_text)
        else:
            self.layout.colour = self.inactive_color
            self.update(self.inactive_text)



import argparse
import glob
import os
import shutil
import subprocess
import sys
from collections.abc import Iterator
from contextlib import contextmanager
from pathlib import Path

from cffi import FFI
from setuptools import Distribution
from setuptools.command.build_ext import build_ext

QW_PATH = (Path(__file__).parent / ".." / "qw").resolve()

PKG_CONFIG = os.environ.get("PKG_CONFIG", "pkg-config")
WAYLAND_SCANNER = os.environ.get("QTILE_WAYLAND_SCANNER", shutil.which("wayland-scanner"))
if not WAYLAND_SCANNER:
    print(
        "Didn't find a wayland-scanner executable in $PATH, and "
        "$QTILE_WAYLAND_SCANNER is not set. Trying to get the path to it from "
        "pkg-config",
        file=sys.stderr,
    )
    WAYLAND_SCANNER = subprocess.run(
        [PKG_CONFIG, "--variable=wayland_scanner", "wayland-scanner"],
        text=True,
        stdout=subprocess.PIPE,
    ).stdout.strip()
WAYLAND_PROTOCOLS = subprocess.run(
    [PKG_CONFIG, "--variable=pkgdatadir", "wayland-protocols"],
    text=True,
    stdout=subprocess.PIPE,
).stdout.strip()

QW_PROTO_IN_PATH = (QW_PATH / ".." / "proto").resolve()

PROTOS = [
    [
        "wlr-layer-shell-unstable-v1-protocol.h",
        f"{QW_PROTO_IN_PATH}/wlr-layer-shell-unstable-v1.xml",
    ],
    [
        "wlr-output-power-management-unstable-v1-protocol.h",
        f"{QW_PROTO_IN_PATH}/wlr-output-power-management-unstable-v1.xml",
    ],
    ["xdg-shell-protocol.h", f"{WAYLAND_PROTOCOLS}/stable/xdg-shell/xdg-shell.xml"],
    [
        "pointer-constraints-unstable-v1-protocol.h",
        f"{WAYLAND_PROTOCOLS}/unstable/pointer-constraints/pointer-constraints-unstable-v1.xml",
    ],
]

QW_PROTO_OUT_PATH = QW_PATH / "proto"
QW_PROTO_OUT_PATH.mkdir(exist_ok=True)

for proto in PROTOS:
    subprocess.run(
        [WAYLAND_SCANNER, "server-header", proto[1], QW_PROTO_OUT_PATH / proto[0]],
        text=True,
        stdout=subprocess.PIPE,
    ).stdout.strip()

CDEF = """
// logging
enum wlr_log_importance {
    WLR_SILENT,
    WLR_ERROR,
    WLR_INFO,
    WLR_DEBUG,
    ...
};
enum wlr_log_importance wlr_log_get_verbosity(void);
extern "Python" void log_cb(enum wlr_log_importance importance,
                                       const char *log_str);

enum wlr_input_device_type {
    WLR_INPUT_DEVICE_KEYBOARD,
    WLR_INPUT_DEVICE_POINTER,
    WLR_INPUT_DEVICE_TOUCH,
    WLR_INPUT_DEVICE_TABLET,
    WLR_INPUT_DEVICE_TABLET_PAD,
    WLR_INPUT_DEVICE_SWITCH,
};

enum libinput_config_accel_profile {
    LIBINPUT_CONFIG_ACCEL_PROFILE_NONE = 0,
    LIBINPUT_CONFIG_ACCEL_PROFILE_FLAT = (1 << 0),
    LIBINPUT_CONFIG_ACCEL_PROFILE_ADAPTIVE = (1 << 1),
};

enum libinput_config_click_method {
        LIBINPUT_CONFIG_CLICK_METHOD_NONE = 0,
        LIBINPUT_CONFIG_CLICK_METHOD_BUTTON_AREAS = (1 << 0),
        LIBINPUT_CONFIG_CLICK_METHOD_CLICKFINGER = (1 << 1),
};

enum libinput_config_tap_button_map {
        LIBINPUT_CONFIG_TAP_MAP_LRM,
        LIBINPUT_CONFIG_TAP_MAP_LMR,
};

enum libinput_config_scroll_method {
        LIBINPUT_CONFIG_SCROLL_NO_SCROLL = 0,
        LIBINPUT_CONFIG_SCROLL_2FG = (1 << 0),
        LIBINPUT_CONFIG_SCROLL_EDGE = (1 << 1),
        LIBINPUT_CONFIG_SCROLL_ON_BUTTON_DOWN = (1 << 2),
};

struct wlr_box {
	int x, y;
	int width, height;
};

struct wlr_cursor {
    struct wlr_cursor_state *state;
    double x, y;
    ...;
};

struct wlr_output {
	char *name;
	char *make;
	char *model;
	char *serial;
	...;
};

// main callbacks
typedef uint32_t xkb_keysym_t;
typedef struct _cairo_surface cairo_surface_t;
extern "Python" void view_activation_cb(struct qw_view *view, void *userdata);
extern "Python" int keyboard_key_cb(xkb_keysym_t, uint32_t, void *userdata);
extern "Python" void manage_view_cb(struct qw_view *view, void *userdata);
extern "Python" void unmanage_view_cb(struct qw_view *view, void *userdata);
extern "Python" void cursor_motion_cb(void *userdata);
extern "Python" int cursor_button_cb(int button, uint32_t mask, bool pressed, int x, int y, void *userdata);
extern "Python" void on_screen_change_cb(void *userdata);
extern "Python" void on_screen_reserve_space_cb(struct qw_output *output, void *userdata);
extern "Python" void on_input_device_added_cb(void *userdata);
extern "Python" bool focus_current_window_cb(void *userdata);
extern "Python" void on_session_lock_cb(bool locked, void *userdata);
extern "Python" struct wlr_box get_current_output_dims_cb(void *userdata);
extern "Python" bool add_idle_inhibitor_cb(void *userdata, void *inhibitor, void *view, bool is_layer_surface, bool is_session_lock_surface);
extern "Python" bool remove_idle_inhibitor_cb(void *userdata, void *inhibitor);
extern "Python" bool check_inhibited_cb(void *userdata);
extern "Python" struct qw_qtile_config *get_qtile_config_cb(void *userdata);
extern "Python" void idle_state_change_cb(void *userdata, int seconds, bool is_idle);

extern "Python" int request_focus_cb(void *userdata);
extern "Python" int request_close_cb(void *userdata);
extern "Python" int request_fullscreen_cb(bool fullscreen, void *userdata);
extern "Python" int request_maximize_cb(bool maximize, void *userdata);
extern "Python" int request_minimize_cb(bool minimize, void *userdata);
extern "Python" void set_title_cb(char* title, void *userdata);
extern "Python" void set_app_id_cb(char* app_id, void *userdata);
"""

cdef_files = [
    "log.h",
    "session-lock.h",
    "server.h",
    "view.h",
    "util.h",
    "output.h",
    "internal-view.h",
    "cursor.h",
    "input-device.h",
    "keyboard.h",
]

for file in cdef_files:
    with open(QW_PATH / file) as f:
        in_private_data = False
        for line in f.readlines():
            if line.startswith("#"):
                continue
            if line.strip().lower().startswith("// private data"):
                in_private_data = True
                CDEF += "    ...;\n"
                continue
            if line.startswith("};") and in_private_data:
                in_private_data = False
            if in_private_data:
                continue
            CDEF += line

SOURCE = "\n".join(f'#include "{header}"' for header in cdef_files) + "\n"


def get_include_path(lib: str) -> str:
    return subprocess.run(
        ["pkg-config", "--variable=includedir", lib], text=True, stdout=subprocess.PIPE
    ).stdout.strip()


INCLUDE_DIRS = [
    os.getenv("QTILE_CAIRO_PATH", "/usr/include/cairo"),
    os.getenv("QTILE_PIXMAN_PATH", "/usr/include/pixman-1"),
    os.getenv("QTILE_LIBDRM_PATH", "/usr/include/libdrm"),
    os.getenv("QTILE_WLROOTS_PATH", "/usr/include/wlroots-0.19"),
    QW_PATH,
    QW_PROTO_OUT_PATH,
]
LIBRARIES = ["wlroots-0.19", "wayland-server", "input", "cairo"]
SOURCE_FILES = glob.glob(f"{QW_PATH}/*.c")
OBJECTS = [Path(src).parent / "build" / Path(src).with_suffix(".o").name for src in SOURCE_FILES]


@contextmanager
def chdir(path: Path) -> Iterator[None]:
    prev_cwd = os.getcwd()
    try:
        os.chdir(path)
        yield
    finally:
        os.chdir(prev_cwd)


def build_objects(debug: bool = False, asan: bool = False) -> None:
    # We have to use relative paths here for output_dir to work as expected
    with chdir(QW_PATH):
        dist = Distribution()
        cmd = build_ext(dist)
        cmd.setup_shlib_compiler()

        extra_preargs = []
        if debug or asan:
            extra_preargs = ["-g3", "-Og"]
        else:
            extra_preargs = ["-O2"]
        extra_preargs.extend(["-fPIC", "-Wall", "-Wextra"])

        cmd.shlib_compiler.compile(
            [os.path.basename(path) for path in SOURCE_FILES],
            output_dir="build",
            macros=[("WLR_USE_UNSTABLE", None)],
            include_dirs=INCLUDE_DIRS,
            extra_preargs=extra_preargs,
        )


def ffi_compile(verbose: bool = False, debug: bool = False, asan: bool = False) -> None:
    # The ffi source of "libqtile.backend.wayland._ffi" means that we'll compile the library file
    # at libqtile/backend/wayland/_ffi.so.
    # This is built at the path specified in tmpdir which is "." by default. So, if we're in a
    # virtual environment, libqtile might be at .venv/lib/python3.13/site-packages/ but if we run
    # the builder script, we'll get a new libqtile folder in th current directory that only has that
    # tree shown above and the libtile library won't find the `_ffi` library.
    # We therefore set the tmpdir to be the path to the folder containing libqtile so the library
    # is always created in the correct folder.
    # The compile command is nested in a function to ensure that the tmpdir value is not overwritten.
    build_objects(debug=debug, asan=asan)

    extra_compile_args = []
    extra_link_args = []
    if debug or asan:
        extra_compile_args.extend(["-g3", "-Og"])
    if asan:
        extra_compile_args.extend(
            ["-fsanitize=address,leak,undefined", "-fno-omit-frame-pointer"]
        )
        extra_link_args.extend(["-fsanitize=address,undefined"])

    ffi = FFI()
    ffi.set_source(
        "libqtile.backend.wayland._ffi",
        SOURCE,
        define_macros=[("WLR_USE_UNSTABLE", None)],
        include_dirs=INCLUDE_DIRS,
        extra_objects=OBJECTS,
        libraries=LIBRARIES,
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
    )
    ffi.cdef(CDEF)
    ffi.compile(
        tmpdir=Path(__file__).parent.parent.parent.parent.parent.as_posix(), verbose=verbose
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Qtile wayland backend build script.")
    parser.add_argument("--debug", action="store_true", help="Add debugging symbols.")
    parser.add_argument(
        "--asan", action="store_true", help="Compile with address sanitisation support."
    )
    args = parser.parse_args()
    ffi_compile(debug=args.debug, asan=args.asan)



import re

from libqtile.widget.generic_poll_text import GenPollCommand


class CapsNumLockIndicator(GenPollCommand):
    """Really simple widget to show the current Caps/Num Lock state."""

    defaults = [("update_interval", 0.5, "Update Time in seconds.")]

    def __init__(self, **config):
        config["cmd"] = ["xset", "q"]
        GenPollCommand.__init__(self, **config)
        self.add_defaults(CapsNumLockIndicator.defaults)

    def parse(self, raw: str):
        if raw.startswith("Keyboard"):
            indicators = re.findall(r"(Caps|Num)\s+Lock:\s*(\w*)", raw)
            return " ".join([" ".join(indicator) for indicator in indicators])
        return ""



from __future__ import annotations

from typing import TYPE_CHECKING

from libqtile.command.base import expose_command
from libqtile.layout.base import Layout, _ClientList
from libqtile.log_utils import logger

if TYPE_CHECKING:
    from typing import Any, Self

    from libqtile.backend.base import Window
    from libqtile.config import ScreenRect
    from libqtile.group import _Group


class _Column(_ClientList):
    # shortcuts for current client and index used in Columns layout
    cw = _ClientList.current_client
    current = _ClientList.current_index

    def __init__(self, split, insert_position, width=100):
        _ClientList.__init__(self)
        self.width = width
        self.split = split
        self.insert_position = insert_position
        self.heights = {}

    @expose_command()
    def info(self) -> dict[str, Any]:
        info = _ClientList.info(self)
        info.update(
            dict(
                heights=[self.heights[c] for c in self.clients],
                split=self.split,
            )
        )
        return info

    def toggle_split(self):
        self.split = not self.split

    def add_client(self, client, height=100):
        _ClientList.add_client(self, client, self.insert_position)
        self.heights[client] = height
        delta = 100 - height
        if delta != 0:
            n = len(self)
            growth = [int(delta / n)] * n
            growth[0] += delta - sum(growth)
            for c, g in zip(self, growth):
                self.heights[c] += g

    def remove(self, client: Window) -> None:
        _ClientList.remove(self, client)
        delta = self.heights[client] - 100
        del self.heights[client]
        if delta != 0:
            n = len(self)
            growth = [int(delta / n)] * n
            growth[0] += delta - sum(growth)
            for c, g in zip(self, growth):
                self.heights[c] += g

    def __str__(self):
        cur = self.current
        return "_Column: " + ", ".join(
            [
                f"[{c.name!s}: {self.heights[c]:d}]"
                if c == cur
                else f"{c.name!s}: {self.heights[c]:d}"
                for c in self.clients
            ]
        )


class Columns(Layout):
    """Extension of the Stack layout.

    The screen is split into columns, which can be dynamically added or
    removed.  Each column can present its windows in 2 modes: split or
    stacked.  In split mode, all windows are presented simultaneously,
    spliting the column space.  In stacked mode, only a single window is
    presented from the stack of windows.  Columns and windows can be
    resized and windows can be shuffled around.

    This layout can also emulate wmii's default layout via:

        layout.Columns(num_columns=1, insert_position=1)

    Or the "Vertical", and "Max", depending on the default parameters.

    An example key configuration is::

        Key([mod], "j", lazy.layout.down()),
        Key([mod], "k", lazy.layout.up()),
        Key([mod], "h", lazy.layout.left()),
        Key([mod], "l", lazy.layout.right()),
        Key([mod, "shift"], "j", lazy.layout.shuffle_down()),
        Key([mod, "shift"], "k", lazy.layout.shuffle_up()),
        Key([mod, "shift"], "h", lazy.layout.shuffle_left()),
        Key([mod, "shift"], "l", lazy.layout.shuffle_right()),
        Key([mod, "control"], "j", lazy.layout.grow_down()),
        Key([mod, "control"], "k", lazy.layout.grow_up()),
        Key([mod, "control"], "h", lazy.layout.grow_left()),
        Key([mod, "control"], "l", lazy.layout.grow_right()),
        Key([mod, "shift", "control"], "h", lazy.layout.swap_column_left()),
        Key([mod, "shift", "control"], "l", lazy.layout.swap_column_right()),
        Key([mod], "Return", lazy.layout.toggle_split()),
        Key([mod], "n", lazy.layout.normalize()),
    """

    _left = 0
    _right = 1

    defaults = [
        ("border_focus", "#881111", "Border colour(s) for the focused window."),
        ("border_normal", "#220000", "Border colour(s) for un-focused windows."),
        (
            "border_focus_stack",
            "#881111",
            "Border colour(s) for the focused window in stacked columns.",
        ),
        (
            "border_normal_stack",
            "#220000",
            "Border colour(s) for un-focused windows in stacked columns.",
        ),
        ("border_width", 2, "Border width."),
        ("single_border_width", None, "Border width for single window."),
        ("border_on_single", False, "Draw a border when there is one only window."),
        ("margin", 0, "Margin of the layout (int or list of ints [N E S W])."),
        (
            "margin_on_single",
            None,
            "Margin when only one window. (int or list of ints [N E S W])",
        ),
        ("split", True, "New columns presentation mode."),
        ("num_columns", 2, "Preferred number of columns."),
        ("grow_amount", 10, "Amount by which to grow a window/column."),
        ("fair", False, "Add new windows to the column with least windows."),
        (
            "insert_position",
            0,
            "Position relative to the current window where new ones are inserted "
            "(0 means right above the current window, 1 means right after).",
        ),
        ("wrap_focus_columns", True, "Wrap the screen when moving focus across columns."),
        ("wrap_focus_rows", True, "Wrap the screen when moving focus across rows."),
        ("wrap_focus_stacks", True, "Wrap the screen when moving focus across stacked."),
        (
            "align",
            _right,
            "Which side of screen new windows will be added to "
            "(one of ``Columns._left`` or ``Columns._right``). "
            "Ignored if 'fair=True'.",
        ),
        ("initial_ratio", 1, "Ratio of first column to second column."),
    ]

    def __init__(self, **config):
        Layout.__init__(self, **config)
        self.add_defaults(Columns.defaults)
        if not self.border_on_single:
            self.single_border_width = 0
        elif self.single_border_width is None:
            self.single_border_width = self.border_width

        if self.margin_on_single is None:
            self.margin_on_single = self.margin
        self.columns = [_Column(self.split, self.insert_position)]
        self.current = 0
        if self.align not in (Columns._left, Columns._right):
            logger.warning(
                "Unexpected value for `align`. Must be Columns._left or Columns._right."
            )
            self.align = Columns._right

    def swap(self, c1: Window, c2: Window) -> None:
        col_c1: tuple[_Column, int]
        col_c2: tuple[_Column, int]

        for c in self.columns:
            for i, w in enumerate(c.clients):
                if w is c1:
                    col_c1 = (c, i)
                elif w is c2:
                    col_c2 = (c, i)

        col_c1[0].clients[col_c1[1]], col_c2[0].clients[col_c2[1]] = c2, c1

        height_c1 = col_c1[0].heights.pop(c1)
        height_c2 = col_c2[0].heights.pop(c2)

        col_c1[0].heights[c2], col_c2[0].heights[c1] = height_c1, height_c2

        self.group.layout_all()

    def clone(self, group: _Group) -> Self:
        c = Layout.clone(self, group)
        c.columns = [_Column(self.split, self.insert_position)]
        return c

    def get_windows(self):
        clients = []
        for c in self.columns:
            clients.extend(c.clients)
        return clients

    @expose_command()
    def info(self) -> dict[str, Any]:
        d = Layout.info(self)
        d["clients"] = []
        d["columns"] = []
        for c in self.columns:
            cinfo = c.info()
            d["clients"].extend(cinfo["clients"])
            d["columns"].append(cinfo)
        d["current"] = self.current
        return d

    def focus(self, client: Window) -> None:
        for i, c in enumerate(self.columns):
            if client in c:
                c.focus(client)
                self.current = i
                break

    @property
    def cc(self):
        return self.columns[self.current]

    def get_ratio_widths(self):
        # Total width is 200
        # main + secondary = 200
        # main = secondary * ratio
        # secondary column is therefore 200 / (1 + ratio)
        # main column is 200 - secondary column
        secondary = 200 // (1 + self.initial_ratio)
        main = 200 - secondary
        return main, secondary

    def add_column(self, prepend=False):
        c = _Column(self.split, self.insert_position)
        if prepend:
            self.columns.insert(0, c)
            self.current += 1
        else:
            self.columns.append(c)
        if len(self.columns) == 2 and not self.fair:
            main, secondary = self.get_ratio_widths()
            self.cc.width = main
            c.width = secondary
        return c

    def remove_column(self, col):
        if len(self.columns) == 1:
            logger.warning("Trying to remove all columns.")
            return
        idx = self.columns.index(col)
        del self.columns[idx]
        if idx <= self.current:
            self.current = max(0, self.current - 1)
        delta = col.width - 100
        if delta != 0:
            n = len(self.columns)
            growth = [int(delta / n)] * n
            growth[0] += delta - sum(growth)
            for c, g in zip(self.columns, growth):
                c.width += g

    def add_client(self, client: Window) -> None:
        c = self.cc
        if len(c) > 0 and len(self.columns) < self.num_columns:
            prepend = self.align is Columns._left
            c = self.add_column(prepend=prepend)
        if self.fair:
            least = min(self.columns, key=len)
            if len(least) < len(c):
                c = least
        self.current = self.columns.index(c)
        c.add_client(client)

    def remove(self, client):
        remove = None
        for c in self.columns:
            if client in c:
                c.remove(client)
                if len(c) == 0 and len(self.columns) > 1:
                    remove = c
                break
        if remove is not None:
            self.remove_column(c)
        return self.columns[self.current].cw

    def configure(self, client: Window, screen_rect: ScreenRect) -> None:
        pos = 0
        for col in self.columns:
            if client in col:
                break
            pos += col.width
        else:
            client.hide()
            return

        if client.has_focus:
            color = self.border_focus if col.split else self.border_focus_stack
        else:
            color = self.border_normal if col.split else self.border_normal_stack

        is_single = len(self.columns) == 1 and (len(col) == 1 or not col.split)
        border = self.single_border_width if is_single else self.border_width
        margin_size = self.margin_on_single if is_single else self.margin

        width = int(0.5 + col.width * screen_rect.width * 0.01 / len(self.columns))
        x = screen_rect.x + int(0.5 + pos * screen_rect.width * 0.01 / len(self.columns))

        if col.split:
            pos = 0
            for c in col:
                if client == c:
                    break
                pos += col.heights[c]
            height = int(0.5 + col.heights[client] * screen_rect.height * 0.01 / len(col))
            y = screen_rect.y + int(0.5 + pos * screen_rect.height * 0.01 / len(col))
            client.place(
                x, y, width - 2 * border, height - 2 * border, border, color, margin=margin_size
            )
            client.unhide()
        elif client == col.cw:
            client.place(
                x,
                screen_rect.y,
                width - 2 * border,
                screen_rect.height - 2 * border,
                border,
                color,
                margin=margin_size,
            )
            client.unhide()
        else:
            client.hide()

    def focus_first(self) -> Window | None:
        """Returns first client in first column of layout"""
        if self.columns:
            return self.columns[0].focus_first()
        return None

    def focus_last(self) -> Window | None:
        """Returns last client in last column of layout"""
        if self.columns:
            return self.columns[-1].focus_last()
        return None

    def focus_next(self, win: Window) -> None:
        """Returns the next client after 'win' in layout,
        or None if there is no such client"""
        # First: try to get next window in column of win (self.columns is non-empty)
        # pylint: disable=undefined-loop-variable
        for idx, col in enumerate(self.columns):
            if win in col:
                if nxt := col.focus_next(win):
                    return nxt
                break
        # if there was no next, get first client from next column
        if idx + 1 < len(self.columns):
            return self.columns[idx + 1].focus_first()
        return None

    def focus_previous(self, win: Window) -> Window | None:
        """Returns the client previous to 'win' in layout.
        or None if there is no such client"""
        # First: try to focus previous client in column (self.columns is non-empty)
        # pylint: disable=undefined-loop-variable
        for idx, col in enumerate(self.columns):
            if win in col:
                if prev := col.focus_previous(win):
                    return prev
                break
        # If there was no previous, get last from previous column
        if idx > 0:
            return self.columns[idx - 1].focus_last()
        return None

    @expose_command()
    def toggle_split(self):
        self.cc.toggle_split()
        self.group.layout_all()

    @expose_command()
    def left(self):
        if self.wrap_focus_columns:
            if len(self.columns) > 1:
                self.current = (self.current - 1) % len(self.columns)
        else:
            if self.current > 0:
                self.current = self.current - 1
        self.group.focus(self.cc.cw, True)

    @expose_command()
    def right(self):
        if self.wrap_focus_columns:
            if len(self.columns) > 1:
                self.current = (self.current + 1) % len(self.columns)
        else:
            if len(self.columns) - 1 > self.current:
                self.current = self.current + 1
        self.group.focus(self.cc.cw, True)

    def want_wrap(self, col):
        if col.split:
            return self.wrap_focus_rows
        return self.wrap_focus_stacks

    @expose_command()
    def up(self):
        col = self.cc
        if self.want_wrap(col):
            if len(col) > 1:
                col.current_index -= 1
        else:
            if col.current_index > 0:
                col.current_index -= 1
        self.group.focus(col.cw, True)

    @expose_command()
    def down(self):
        col = self.cc
        if self.want_wrap(col):
            if len(col) > 1:
                col.current_index += 1
        else:
            if col.current_index < len(col) - 1:
                col.current_index += 1
        self.group.focus(col.cw, True)

    @expose_command()
    def next(self) -> None:
        if self.cc.split and self.cc.current < len(self.cc) - 1:
            self.cc.current += 1
        elif self.columns:
            self.current = (self.current + 1) % len(self.columns)
            if self.cc.split:
                self.cc.current = 0
        self.group.focus(self.cc.cw, True)

    @expose_command()
    def previous(self) -> None:
        if self.cc.split and self.cc.current > 0:
            self.cc.current -= 1
        elif self.columns:
            self.current = (self.current - 1) % len(self.columns)
            if self.cc.split:
                self.cc.current = len(self.cc) - 1
        self.group.focus(self.cc.cw, True)

    @expose_command()
    def shuffle_left(self):
        cur = self.cc
        client = cur.cw
        if client is None:
            return
        if self.current > 0:
            self.current -= 1
            new = self.cc
            new.add_client(client, cur.heights[client])
            cur.remove(client)
            if len(cur) == 0:
                self.remove_column(cur)
        elif len(cur) > 1:
            new = self.add_column(True)
            new.add_client(client, cur.heights[client])
            cur.remove(client)
            self.current = 0
        else:
            return
        self.group.layout_all()

    @expose_command()
    def shuffle_right(self):
        cur = self.cc
        client = cur.cw
        if client is None:
            return
        if self.current + 1 < len(self.columns):
            self.current += 1
            new = self.cc
            new.add_client(client, cur.heights[client])
            cur.remove(client)
            if len(cur) == 0:
                self.remove_column(cur)
        elif len(cur) > 1:
            new = self.add_column()
            new.add_client(client, cur.heights[client])
            cur.remove(client)
            self.current = len(self.columns) - 1
        else:
            return
        self.group.layout_all()

    @expose_command()
    def shuffle_up(self):
        if self.cc.current_index > 0:
            self.cc.shuffle_up()
            self.group.layout_all()

    @expose_command()
    def shuffle_down(self):
        if self.cc.current_index + 1 < len(self.cc):
            self.cc.shuffle_down()
            self.group.layout_all()

    @expose_command()
    def grow_left(self):
        if self.current > 0:
            if self.columns[self.current - 1].width > self.grow_amount:
                self.columns[self.current - 1].width -= self.grow_amount
                self.cc.width += self.grow_amount
                self.group.layout_all()
        elif len(self.columns) > 1:
            if self.columns[0].width > self.grow_amount:
                self.columns[1].width += self.grow_amount
                self.cc.width -= self.grow_amount
                self.group.layout_all()

    @expose_command()
    def grow_right(self):
        if self.current + 1 < len(self.columns):
            if self.columns[self.current + 1].width > self.grow_amount:
                self.columns[self.current + 1].width -= self.grow_amount
                self.cc.width += self.grow_amount
                self.group.layout_all()
        elif len(self.columns) > 1:
            if self.cc.width > self.grow_amount:
                self.cc.width -= self.grow_amount
                self.columns[self.current - 1].width += self.grow_amount
                self.group.layout_all()

    @expose_command()
    def grow_up(self):
        col = self.cc
        if col.current > 0:
            if col.heights[col[col.current - 1]] > self.grow_amount:
                col.heights[col[col.current - 1]] -= self.grow_amount
                col.heights[col.cw] += self.grow_amount
                self.group.layout_all()
        elif len(col) > 1:
            if col.heights[col.cw] > self.grow_amount:
                col.heights[col[1]] += self.grow_amount
                col.heights[col.cw] -= self.grow_amount
                self.group.layout_all()

    @expose_command()
    def grow_down(self):
        col = self.cc
        if col.current + 1 < len(col):
            if col.heights[col[col.current + 1]] > self.grow_amount:
                col.heights[col[col.current + 1]] -= self.grow_amount
                col.heights[col.cw] += self.grow_amount
                self.group.layout_all()
        elif len(col) > 1:
            if col.heights[col.cw] > self.grow_amount:
                col.heights[col[col.current - 1]] += self.grow_amount
                col.heights[col.cw] -= self.grow_amount
                self.group.layout_all()

    @expose_command()
    def normalize(self):
        """Give columns equal widths."""
        for col in self.columns:
            for client in col:
                col.heights[client] = 100
            col.width = 100
        self.group.layout_all()

    @expose_command()
    def reset(self):
        """Resets column widths, respecting 'initial_ratio' value."""
        if self.initial_ratio == 1 or len(self.columns) == 1 or self.fair:
            self.normalize()
            return

        main, secondary = self.get_ratio_widths()

        if self.align == Columns._right:
            self.columns[0].width = main
            self.columns[1].width = secondary
        else:
            self.columns[-1].width = main
            self.columns[-2].width = secondary

        self.group.layout_all()

    def swap_column(self, src, dst):
        self.columns[src], self.columns[dst] = self.columns[dst], self.columns[src]
        self.current = dst
        self.group.layout_all()

    @expose_command()
    def swap_column_left(self):
        src = self.current
        dst = src - 1 if src > 0 else len(self.columns) - 1
        self.swap_column(src, dst)

    @expose_command()
    def swap_column_right(self):
        src = self.current
        dst = src + 1 if src < len(self.columns) - 1 else 0
        self.swap_column(src, dst)



from datetime import UTC, datetime

from libqtile.confreader import ConfigError
from libqtile.widget import base
from libqtile.widget.clock import Clock


class VerticalClock(Clock):
    """
    A simple but flexible text-based clock for vertical bars.

    Unlike the ``Clock`` widget, ``VerticalClock`` will display text horizontally in the bar.
    """

    orientations = base.ORIENTATION_VERTICAL

    defaults = [
        (
            "format",
            ["%H", "%M"],
            "A list of Python datetime format string. Each string is printed as a separate line.",
        ),
        (
            "foreground",
            "fff",
            "Text colour. A single string will be applied to all fields. "
            "Alternatively, users can provide a list of strings with each colour being applied to the corresponding text format.",
        ),
        (
            "fontsize",
            None,
            "Font size. A single value will be applied to all fields. "
            "Alternatively, users can provide a list of sizes with each size being applied to the corresponding text format.",
        ),
    ]

    def __init__(self, **config):
        Clock.__init__(self, **config)
        self.add_defaults(VerticalClock.defaults)
        self.layouts = []

    def _to_list(self, value):
        return [value] * len(self.format)

    def _configure(self, qtile, bar):
        base._Widget._configure(self, qtile, bar)

        if self.fontsize is None:
            self.fontsize = self._to_list(self.bar.size - self.bar.size / 5)
        elif isinstance(self.fontsize, int):
            self.fontsize = self._to_list(self.fontsize)
        elif not isinstance(self.fontsize, list):
            raise ConfigError("'fontsize' should be an integer or a list of integers.")
        elif not all(isinstance(fontsize, int) for fontsize in self.fontsize):
            raise ConfigError("'fontsize' should be of integers.")
        elif len(self.fontsize) != len(self.format):
            raise ConfigError("'fontsize' list should have same number of items as 'format'.")

        if isinstance(self.foreground, str):
            self.foreground = self._to_list(self.foreground)
        elif not isinstance(self.foreground, list):
            raise ConfigError("'foreground' should be a string or a list of strings.")
        elif not all(isinstance(foreground, str) for foreground in self.foreground):
            raise ConfigError("'foreground' list should be of strings.")
        elif len(self.foreground) != len(self.format):
            raise ConfigError("'foreground' list should have same number of items as 'format'.")

        if self.padding is None:
            self.padding = (sum(self.fontsize) / len(self.fontsize)) // 2

        self.layouts = [
            self.drawer.textlayout(
                self.formatted_text,
                fg,
                self.font,
                size,
                self.fontshadow,
                markup=self.markup,
            )
            for _, fg, size in zip(self.format, self.foreground, self.fontsize)
        ]

    def calculate_length(self):
        return sum(l.height + self.padding for l in self.layouts) + self.padding

    def update(self, time):
        for layout, fmt in zip(self.layouts, self.format):
            layout.text = time.strftime(fmt)
            layout.width = self.bar.size
        self.draw()

    @property
    def can_draw(self):
        return all(layout is not None for layout in self.layouts)

    # adding .5 to get a proper seconds value because glib could
    # theoreticaly call our method too early and we could get something
    # like (x-1).999 instead of x.000
    def poll(self):
        if self.timezone:
            now = datetime.now(UTC).astimezone(self.timezone)
        else:
            now = datetime.now(UTC).astimezone()
        return now + self.DELTA

    def draw(self):
        if not self.can_draw:
            return
        offset = self.padding
        self.drawer.clear(self.background or self.bar.background)

        for layout in self.layouts:
            self.drawer.ctx.save()
            self.drawer.ctx.translate(0, offset)
            layout.draw(0, 0)
            offset += layout.height + self.padding
            self.drawer.ctx.restore()

        self.draw_at_default_position()

    def finalize(self):
        for layout in self.layouts:
            layout.finalize()
            layout = None

        base._Widget.finalize(self)



import libcst as cst
import libcst.matchers as m
from libcst import codemod
from libcst.codemod.visitors import AddImportsVisitor

from libqtile.scripts.migrations._base import (
    Change,
    Check,
    MigrationTransformer,
    NoChange,
    _QtileMigrator,
    add_migration,
)

# Three methods had their names changed beyond removing the prefix
MIGRATION_MAP = {
    "cmd_hints": "get_hints",
    "cmd_groups": "get_groups",
    "cmd_screens": "get_screens",
    "cmd_opacity": "set_opacity",
}


def is_cmd(value):
    return value.startswith("cmd_")


class CmdPrefixTransformer(MigrationTransformer):
    def __init__(self, *args, **kwargs):
        MigrationTransformer.__init__(self, *args, **kwargs)
        self.needs_import = False

    @m.call_if_inside(m.Call())
    @m.leave(m.Name(m.MatchIfTrue(is_cmd)))
    def change_func_call_name(self, original_node, updated_node) -> cst.Name:
        """Removes cmd_prefix from a call where the command begins with the cmd_ prefix."""
        name = original_node.value

        if name in MIGRATION_MAP:
            replacement = MIGRATION_MAP[name]

        else:
            replacement = name[4:]

        self.lint(
            original_node,
            f"Use of 'cmd_' prefix is deprecated. '{name}' should be replaced with '{replacement}'",
        )

        return updated_node.with_changes(value=replacement)

    @m.call_if_inside(m.ClassDef())
    @m.leave(m.FunctionDef(name=m.Name(m.MatchIfTrue(is_cmd))))
    def change_func_def(self, original_node, updated_node) -> cst.FunctionDef:
        """
        Renames method definitions using the cmd_ prefix and adds the
        @expose_command decorator.

        Also sets a flag to show that the script should add the relevant
        import if it's missing.
        """
        decorator = cst.Decorator(cst.Name("expose_command"))
        name = original_node.name
        updated = name.with_changes(value=name.value[4:])

        self.lint(
            original_node,
            "Use of 'cmd_' prefix is deprecated. Use '@expose_command' when defining methods.",
        )

        self.needs_import = True

        return updated_node.with_changes(name=updated, decorators=[decorator])


class RemoveCmdPrefix(_QtileMigrator):
    ID = "RemoveCmdPrefix"

    SUMMARY = "Removes ``cmd_`` prefix from method calls and definitions."

    HELP = """
    The ``cmd_`` prefix was used to identify methods that should be exposed to
    qtile's command API. This has been deprecated and so calls no longer require
    the prefix.

    For example:

    .. code:: python

      qtile.cmd_spawn("vlc")

    would be replaced with:

    .. code:: python

      qtile.spawn("vlc")

    Where users have created their own widgets with methods using this prefix,
    the syntax has also changed:

    For example:

    .. code:: python

        class MyWidget(libqtile.widget.base._Widget):
            def cmd_my_command(self):
                pass

    Should be updated as follows:

    .. code:: python

        from libqtile.command.base import expose_command

        class MyWidget(libqtile.widget.base._Widget):
            @expose_command
            def my_command(self):
                pass
    """

    AFTER_VERSION = "0.22.1"

    TESTS = [
        Change("""qtile.cmd_spawn("alacritty")""", """qtile.spawn("alacritty")"""),
        Change("""qtile.cmd_groups()""", """qtile.get_groups()"""),
        Change("""qtile.cmd_screens()""", """qtile.get_screens()"""),
        Change("""qtile.current_window.cmd_hints()""", """qtile.current_window.get_hints()"""),
        Change(
            """qtile.current_window.cmd_opacity(0.5)""",
            """qtile.current_window.set_opacity(0.5)""",
        ),
        Change(
            """
            class MyWidget(widget.Clock):
                def cmd_my_command(self):
                    pass
            """,
            """
            from libqtile.command.base import expose_command

            class MyWidget(widget.Clock):
                @expose_command
                def my_command(self):
                    pass
            """,
        ),
        NoChange(
            """
            def cmd_some_other_func():
                pass
            """
        ),
        Check(
            """
            from libqtile import qtile, widget

            class MyClock(widget.Clock):
                def cmd_my_exposed_command(self):
                    pass

            def my_func(qtile):
                qtile.cmd_spawn("rickroll")
                hints = qtile.current_window.cmd_hints()
                groups = qtile.cmd_groups()
                screens = qtile.cmd_screens()
                qtile.current_window.cmd_opacity(0.5)

            def cmd_some_other_func():
                pass
            """,
            """
            from libqtile import qtile, widget
            from libqtile.command.base import expose_command

            class MyClock(widget.Clock):
                @expose_command
                def my_exposed_command(self):
                    pass

            def my_func(qtile):
                qtile.spawn("rickroll")
                hints = qtile.current_window.get_hints()
                groups = qtile.get_groups()
                screens = qtile.get_screens()
                qtile.current_window.set_opacity(0.5)

            def cmd_some_other_func():
                pass
            """,
        ),
    ]

    def run(self, original):
        # Run the base migrations
        transformer = CmdPrefixTransformer()
        updated = original.visit(transformer)
        self.update_lint(transformer)

        # Check if we need to add an import line
        if transformer.needs_import:
            # We use the built-in visitor to add the import
            context = codemod.CodemodContext()
            AddImportsVisitor.add_needed_import(
                context, "libqtile.command.base", "expose_command"
            )
            visitor = AddImportsVisitor(context)

            # Run the visitor over the updated code
            updated = updated.visit(visitor)

        return original, updated


add_migration(RemoveCmdPrefix)



from __future__ import annotations

import asyncio
import contextlib
import signal
from typing import TYPE_CHECKING

from libqtile.log_utils import logger

if TYPE_CHECKING:
    from collections.abc import Callable


class LoopContext(contextlib.AbstractAsyncContextManager):
    def __init__(
        self,
        signals: dict[signal.Signals, Callable] | None = None,
    ) -> None:
        super().__init__()
        self._signals = signals or {}
        self._stopped = False

    async def __aenter__(self) -> LoopContext:
        self._stopped = False
        loop = asyncio.get_running_loop()
        loop.set_exception_handler(self._handle_exception)
        for sig, handler in self._signals.items():
            loop.add_signal_handler(sig, handler)

        return self

    async def __aexit__(self, *args) -> None:
        self._stopped = True

        await self._cancel_all_tasks()

        loop = asyncio.get_running_loop()
        map(loop.remove_signal_handler, self._signals.keys())
        loop.set_exception_handler(None)

    async def _cancel_all_tasks(self) -> None:
        # we don't want to cancel this task, so filter all_tasks
        # generator to filter in place
        pending = (task for task in asyncio.all_tasks() if task is not asyncio.current_task())
        for task in pending:
            task.cancel()
            with contextlib.suppress(asyncio.CancelledError):
                await task

    def _handle_exception(
        self,
        loop: asyncio.AbstractEventLoop,
        context: dict,
    ) -> None:
        if "exception" in context:
            exc = context["exception"]
            # CancelledErrors happen when we simply cancel the main task during
            # a normal restart procedure
            if not isinstance(exc, asyncio.CancelledError):
                logger.exception("Exception in event loop:", exc_info=exc)  # noqa: G202
        else:
            logger.error("unhandled error in event loop: %s", context["msg"])



from __future__ import annotations

import xcffib
from xcffib.xproto import ClientMessageData, ClientMessageEvent, EventMask, SetMode

from libqtile import bar
from libqtile.backend.x11 import window
from libqtile.confreader import ConfigError
from libqtile.widget import base

XEMBED_PROTOCOL_VERSION = 0


class Icon(window._Window):
    _window_mask = EventMask.StructureNotify | EventMask.PropertyChange | EventMask.Exposure

    def __init__(self, win, qtile, systray):
        window._Window.__init__(self, win, qtile)
        self.hidden = True
        self.systray = systray
        # win.get_name() may return None when apps provide a temporary window before the icon window
        # we need something in self.name in order to sort icons so we use the window's WID.
        self.name = win.get_name() or str(win.wid)
        self.update_size()
        self._wm_class: list[str] | None = None

    def __eq__(self, other):
        if not isinstance(other, Icon):
            return False

        return self.window.wid == other.window.wid

    def update_size(self):
        icon_size = self.systray.icon_size
        self.update_hints()

        width = self.hints.get("min_width", icon_size)
        height = self.hints.get("min_height", icon_size)

        width = max(width, icon_size)
        height = max(height, icon_size)

        if height > icon_size:
            width = width * icon_size // height
            height = icon_size

        self.width = width
        self.height = height
        return False

    def handle_PropertyNotify(self, e):  # noqa: N802
        name = self.qtile.core.conn.atoms.get_name(e.atom)
        if name == "_XEMBED_INFO":
            info = self.window.get_property("_XEMBED_INFO", unpack=int)
            if info and info[1]:
                self.systray.bar.draw()

        return False

    def handle_DestroyNotify(self, event):  # noqa: N802
        wid = event.window
        icon = self.qtile.windows_map.pop(wid)
        self.systray.tray_icons.remove(icon)
        self.systray.bar.draw()
        return False

    handle_UnmapNotify = handle_DestroyNotify  # noqa: N815


class Systray(base._Widget, window._Window):
    """
    A widget that manages system tray.

    Only one Systray widget is allowed. Adding additional Systray
    widgets will result in a ConfigError.

    .. note::
        Icons will not render correctly where the bar/widget is
        drawn with a semi-transparent background. Instead, icons
        will be drawn with a transparent background.

        If using this widget it is therefore recommended to use
        a fully opaque background colour or a fully transparent
        one.
    """

    _instances = 0

    _window_mask = EventMask.StructureNotify | EventMask.Exposure

    orientations = base.ORIENTATION_BOTH

    supported_backends = {"x11"}

    defaults = [
        ("icon_size", 20, "Icon width"),
        ("padding", 5, "Padding between icons"),
    ]

    def __init__(self, **config):
        base._Widget.__init__(self, bar.CALCULATED, **config)
        self.add_defaults(Systray.defaults)
        self.tray_icons = []
        self.screen = 0
        self._name = config.get("name", "systray")
        self._wm_class: list[str] | None = None

    def calculate_length(self):
        if self.bar.horizontal:
            length = sum(i.width for i in self.tray_icons)
        else:
            length = sum(i.height for i in self.tray_icons)
        length += self.padding * len(self.tray_icons)
        return length

    def _configure(self, qtile, bar):
        base._Widget._configure(self, qtile, bar)

        if self.configured:
            return

        if Systray._instances > 0:
            raise ConfigError("Only one Systray can be used.")

        self.conn = conn = qtile.core.conn
        win = conn.create_window(-1, -1, 1, 1)
        window._Window.__init__(self, window.XWindow(conn, win.wid), qtile)
        qtile.windows_map[win.wid] = self

        # window._Window.__init__ overwrites the widget name so we need to restore it
        self.name = self._name

        # Even when we have multiple "Screen"s, we are setting up as the system
        # tray on a particular X display, that is the screen we need to
        # reference in the atom
        if qtile.current_screen:
            self.screen = qtile.current_screen.index
        self.bar = bar
        atoms = conn.atoms

        # We need tray to tell icons which visual to use.
        # This needs to be the same as the bar/widget.
        # This mainly benefits transparent bars.
        conn.conn.core.ChangeProperty(
            xcffib.xproto.PropMode.Replace,
            win.wid,
            atoms["_NET_SYSTEM_TRAY_VISUAL"],
            xcffib.xproto.Atom.VISUALID,
            32,
            1,
            [self.drawer._visual.visual_id],
        )

        conn.conn.core.SetSelectionOwner(
            win.wid, atoms[f"_NET_SYSTEM_TRAY_S{self.screen:d}"], xcffib.CurrentTime
        )
        data = [
            xcffib.CurrentTime,
            atoms[f"_NET_SYSTEM_TRAY_S{self.screen:d}"],
            win.wid,
            0,
            0,
        ]
        union = ClientMessageData.synthetic(data, "I" * 5)
        event = ClientMessageEvent.synthetic(
            format=32, window=qtile.core._root.wid, type=atoms["MANAGER"], data=union
        )
        qtile.core._root.send_event(event, mask=EventMask.StructureNotify)

        Systray._instances += 1

    def create_mirror(self):
        """
        Systray cannot be mirrored as we do not use a Drawer object to render icons.

        Return new, unconfigured instance so that, when the bar tries to configure it
        again, a ConfigError is raised.
        """
        return Systray()

    def handle_ClientMessage(self, event):  # noqa: N802
        atoms = self.conn.atoms

        opcode = event.type
        data = event.data.data32
        message = data[1]
        wid = data[2]

        parent = self.bar.window.window

        if opcode == atoms["_NET_SYSTEM_TRAY_OPCODE"] and message == 0:
            w = window.XWindow(self.conn, wid)
            icon = Icon(w, self.qtile, self)
            if icon not in self.tray_icons:
                self.tray_icons.append(icon)
                self.tray_icons.sort(key=lambda icon: icon.name)
                self.qtile.windows_map[wid] = icon

            self.conn.conn.core.ChangeSaveSet(SetMode.Insert, wid)
            self.conn.conn.core.ReparentWindow(wid, parent.wid, 0, 0)
            self.conn.conn.flush()

            info = icon.window.get_property("_XEMBED_INFO", unpack=int)

            if not info:
                self.bar.draw()
                return False

            if info[1]:
                self.bar.draw()

        return False

    def draw(self):
        _xoffset = self.offsetx if self.bar.horizontal else self.offsety
        _yoffset = self.offsety if self.bar.horizontal else self.offsetx
        xoffset = _xoffset + self.padding
        yoffset = self.bar.size // 2 - self.icon_size // 2 + _yoffset
        self.drawer.clear(self.background or self.bar.background)
        self.draw_at_default_position()

        for pos, icon in enumerate(self.tray_icons):
            icon.window.set_attribute(backpixmap=self.drawer.pixmap)
            if self.bar.horizontal:
                step = icon.width
                icon.place(xoffset, yoffset, icon.width, self.icon_size, 0, None)
            else:
                step = icon.height
                icon.place(yoffset, xoffset, icon.width, self.icon_size, 0, None)
            xoffset += step + self.padding

            if icon.hidden:
                icon.unhide()
                data = [
                    self.conn.atoms["_XEMBED_EMBEDDED_NOTIFY"],
                    xcffib.xproto.Time.CurrentTime,
                    0,
                    self.bar.window.wid,
                    XEMBED_PROTOCOL_VERSION,
                ]
                u = xcffib.xproto.ClientMessageData.synthetic(data, "I" * 5)
                event = xcffib.xproto.ClientMessageEvent.synthetic(
                    format=32, window=icon.wid, type=self.conn.atoms["_XEMBED"], data=u
                )
                self.window.send_event(event)

    def finalize(self):
        base._Widget.finalize(self)
        atoms = self.conn.atoms

        try:
            self.conn.conn.core.SetSelectionOwner(
                0,
                atoms[f"_NET_SYSTEM_TRAY_S{self.screen:d}"],
                xcffib.CurrentTime,
            )
            self.hide()

            root = self.qtile.core._root.wid
            for icon in self.tray_icons:
                self.conn.conn.core.ReparentWindow(icon.window.wid, root, 0, 0)
            self.conn.conn.flush()

            self.conn.conn.core.DestroyWindow(self.wid)
        except xcffib.ConnectionException:
            self.hidden = True  # Usually set in self.hide()

        del self.qtile.windows_map[self.wid]
        Systray._instances -= 1

    def info(self):
        info = window._Window.info(self)
        info["widget"] = base._Widget.info(self)
        return info



from __future__ import annotations

from typing import TYPE_CHECKING

from libqtile import hook, utils
from libqtile.command.base import CommandObject, expose_command
from libqtile.log_utils import logger

if TYPE_CHECKING:
    from libqtile.command.base import ItemT


class _Group(CommandObject):
    """A container for a bunch of windows

    Analogous to workspaces in other window managers. Each client window
    managed by the window manager belongs to exactly one group.

    A group is identified by its name but displayed in GroupBox widget by its label.
    """

    def __init__(self, name, layout=None, label=None, screen_affinity=None, persist=False):
        self.screen_affinity = screen_affinity
        self.name = name
        self.label = name if label is None else label
        self.custom_layout = layout  # will be set on _configure
        self.windows = []
        self.tiled_windows = set()
        self.qtile = None
        self.layouts = []
        self.floating_layout = None
        # self.focus_history lists the group's windows in the order they
        # received focus, from the oldest (first item) to the currently
        # focused window (last item); NB the list does *not* contain any
        # windows that never received focus; refer to self.windows for the
        # complete set
        self.focus_history = []
        self.screen = None
        self.current_layout = None
        self.last_focused = None
        self.persist = persist

    def _configure(self, layouts, floating_layout, qtile):
        self.screen = None
        self.current_layout = 0
        self.focus_history = []
        self.windows = []
        self.qtile = qtile
        self.layouts = [i.clone(self) for i in layouts]
        self.floating_layout = floating_layout
        if self.custom_layout is not None:
            self.layout = self.custom_layout
            self.custom_layout = None

    @property
    def current_window(self):
        try:
            return self.focus_history[-1]
        except IndexError:
            # no window has focus
            return None

    @current_window.setter
    def current_window(self, win):
        try:
            self.focus_history.remove(win)
        except ValueError:
            # win has never received focus before
            pass
        self.focus_history.append(win)

    def _remove_from_focus_history(self, win):
        try:
            index = self.focus_history.index(win)
        except ValueError:
            # win has never received focus
            return False
        else:
            del self.focus_history[index]
            # return True if win was the last item (i.e. it was current_window)
            return index == len(self.focus_history)

    @property
    def layout(self):
        return self.layouts[self.current_layout]

    @layout.setter
    def layout(self, layout):
        """
        Parameters
        ==========
        layout :
            a string with matching the name of a Layout object.
        """
        for index, obj in enumerate(self.layouts):
            if obj.name == layout:
                self.use_layout(index)
                return
        logger.error("No such layout: %s", layout)

    def use_layout(self, index: int):
        assert -len(self.layouts) <= index < len(self.layouts), "layout index out of bounds"
        self.layout.hide()
        self.current_layout = index % len(self.layouts)
        hook.fire("layout_change", self.layouts[self.current_layout], self)
        self.layout_all()
        if self.screen is not None:
            screen_rect = self.screen.get_rect()
            self.layout.show(screen_rect)

    def use_next_layout(self):
        self.use_layout((self.current_layout + 1) % (len(self.layouts)))

    def use_previous_layout(self):
        self.use_layout((self.current_layout - 1) % (len(self.layouts)))

    def layout_all(self, warp=False, focus=True):
        """Layout the floating layer, then the current layout.

        Parameters
        ==========
        focus :
            If we have have a current_window give it focus, optionally moving warp
            to it.
        """
        if self.screen and self.windows:
            with self.qtile.core.masked():
                normal = [x for x in self.windows if not x.floating]
                floating = [x for x in self.windows if x.floating and not x.minimized]
                screen_rect = self.screen.get_rect()
                if normal:
                    try:
                        self.layout.layout(normal, screen_rect)
                    except Exception:
                        logger.exception("Exception in layout %s", self.layout.name)
                if floating:
                    self.floating_layout.layout(floating, screen_rect)
                if focus:
                    if self.current_window and self.screen == self.qtile.current_screen:
                        self.current_window.focus(warp)
                    else:
                        # Screen has lost focus so we reset record of focused window so
                        # focus will warp when screen is focused again
                        self.last_focused = None
        elif self.screen and not self.windows and self.screen == self.qtile.current_screen:
            # Clear active window when switching to an empty group on the current screen
            self.qtile.core.clear_focus()

    def set_screen(self, screen, warp=True):
        """Set this group's screen to screen"""
        if screen == self.screen:
            return
        self.screen = screen
        if self.screen:
            # move all floating guys offset to new screen
            self.floating_layout.to_screen(self, self.screen)
            self.layout_all(warp=warp and self.qtile.config.cursor_warp)
            screen_rect = self.screen.get_rect()
            self.floating_layout.show(screen_rect)
            self.layout.show(screen_rect)
        else:
            self.hide()

    def hide(self):
        self.screen = None
        with self.qtile.core.masked():
            for i in self.windows:
                i.hide()
            self.layout.hide()

    def focus(self, win, warp=True, force=False):
        """Focus the given window

        If win is in the group, blur any windows and call ``focus`` on the
        layout (in case it wants to track anything), fire focus_change hook and
        invoke layout_all.

        Parameters
        ==========
        win :
            Window to focus
        warp :
            Warp pointer to win. This should basically always be True, unless
            the focus event is coming from something like EnterNotify, where
            the user is actively using the mouse, or on full screen layouts
            where only one window is "maximized" at a time, and it doesn't make
            sense for the mouse to automatically move.
        """
        if self.qtile._drag and not force:
            # don't change focus while dragging windows (unless forced)
            return
        if win:
            if win not in self.windows:
                return

            # ignore focus events if window is the current window
            if win is self.last_focused:
                warp = False

            self.current_window = win
            self.last_focused = self.current_window
            if win.floating:
                for layout in self.layouts:
                    layout.blur()
                self.floating_layout.focus(win)
            else:
                self.floating_layout.blur()
                for layout in self.layouts:
                    layout.focus(win)
            hook.fire("focus_change")
            self.layout_all(warp)

    @expose_command()
    def info(self):
        """Returns a dictionary of info for this group"""
        return dict(
            name=self.name,
            label=self.label,
            focus=self.current_window.name if self.current_window else None,
            tiled_windows={i.name for i in self.tiled_windows},
            windows=[i.name for i in self.windows],
            focus_history=[i.name for i in self.focus_history],
            layout=self.layout.name,
            layouts=[i.name for i in self.layouts],
            floating_info=self.floating_layout.info(),
            screen=self.screen.index if self.screen else None,
        )

    def add(self, win, force=False):
        hook.fire("group_window_add", self, win)
        if win not in self.windows:
            self.windows.append(win)
        win.group = self
        if self.qtile.config.auto_fullscreen and win.wants_to_fullscreen:
            win.fullscreen = True
        elif self.floating_layout.match(win) and not win.fullscreen:
            win.floating = True
        if win.floating and not win.fullscreen:
            self.floating_layout.add_client(win)
        else:
            self.tiled_windows.add(win)
            for i in self.layouts:
                i.add_client(win)
        if win.can_steal_focus:
            self.focus(win, warp=True, force=force)
        else:
            self.layout_all(focus=False)

    def remove(self, win, force=False):
        hook.fire("group_window_remove", self, win)

        if self.qtile.config.focus_previous_on_window_remove:
            index = self.focus_history.index(win)
            try:
                previous_win = self.focus_history[index - 1]
            except IndexError:
                previous_win = None
            if previous_win not in self.windows:
                previous_win = None
        else:
            previous_win = None

        self.windows.remove(win)
        hadfocus = self._remove_from_focus_history(win)
        win.group = None

        if win.floating:
            nextfocus = self.floating_layout.remove(win)

            nextfocus = (
                previous_win
                or nextfocus
                or self.current_window
                or self.layout.focus_first()
                or self.floating_layout.focus_first(group=self)
            )
        # Remove from the tiled layouts if it was not floating or fullscreen
        if not win.floating or win.fullscreen:
            for i in self.layouts:
                if i is self.layout:
                    nextfocus = i.remove(win)
                else:
                    i.remove(win)

            nextfocus = (
                previous_win
                or nextfocus
                or self.floating_layout.focus_first(group=self)
                or self.current_window
                or self.layout.focus_first()
            )

            if win in self.tiled_windows:
                self.tiled_windows.remove(win)

        # a notification may not have focus
        if hadfocus:
            self.focus(nextfocus, warp=True, force=force)
            # no next focus window means focus changed to nothing
            if not nextfocus:
                hook.fire("focus_change")
        elif self.screen:
            self.layout_all()

    def mark_floating(self, win, floating):
        if floating:
            if win in self.floating_layout.find_clients(self):
                # already floating
                pass
            else:
                # Remove from the tiled windows list if the window is not fullscreen
                if not win.fullscreen:
                    self.tiled_windows.remove(win)
                    # Remove the window from the layout if it is not fullscreen
                    for i in self.layouts:
                        i.remove(win)
                        if win is self.current_window:
                            i.blur()
                    self.floating_layout.add_client(win)
                    if win is self.current_window:
                        self.floating_layout.focus(win)
        else:
            self.floating_layout.remove(win)
            self.floating_layout.blur()
            # A window that was fullscreen should only be added if it was not a tiled window
            if win not in self.tiled_windows:
                for i in self.layouts:
                    i.add_client(win)
                self.tiled_windows.add(win)
            if win is self.current_window:
                for i in self.layouts:
                    i.focus(win)
        self.layout_all()

    def _items(self, name) -> ItemT:
        if name == "layout":
            return True, list(range(len(self.layouts)))
        if name == "screen" and self.screen is not None:
            return True, []
        if name == "window":
            return self.current_window is not None, [i.wid for i in self.windows]
        return None

    def _select(self, name, sel):
        if name == "layout":
            if sel is None:
                return self.layout
            return utils.lget(self.layouts, sel)
        if name == "screen":
            return self.screen
        if name == "window":
            if sel is None:
                return self.current_window
            for i in self.windows:
                if i.wid == sel:
                    return i
        raise RuntimeError(f"Invalid selection: {name}")

    @expose_command()
    def setlayout(self, layout):
        self.layout = layout

    @expose_command()
    def toscreen(self, screen=None, toggle=False):
        """Pull a group to a specified screen.

        Parameters
        ==========
        screen :
            Screen offset. If not specified, we assume the current screen.
        toggle :
            If this group is already on the screen, then the group is toggled
            with last used

        Examples
        ========
        Pull group to the current screen::

            toscreen()

        Pull group to screen 0::

            toscreen(0)
        """
        if screen is None:
            screen = self.qtile.current_screen
        else:
            screen = self.qtile.screens[screen]

        if screen.group == self:
            if toggle:
                screen.toggle_group(self)
        else:
            screen.set_group(self)

        self.qtile.core.check_screen_fullscreen_background(screen)

    def _get_group(self, direction, skip_empty=False, skip_managed=False):
        """Find a group walking the groups list in the specified direction

        Parameters
        ==========
        skip_empty :
            skips the empty groups
        skip_managed :
            skips the groups that have a screen
        """

        def match(group):
            from libqtile import scratchpad

            if group is self:
                return True
            if skip_empty and not group.windows:
                return False
            if skip_managed and group.screen:
                return False
            if isinstance(group, scratchpad.ScratchPad):
                return False
            return True

        try:
            groups = [group for group in self.qtile.groups if match(group)]
            index = (groups.index(self) + direction) % len(groups)
            return groups[index]
        except ValueError:
            # group is not managed
            return None

    def get_previous_group(self, skip_empty=False, skip_managed=False):
        return self._get_group(-1, skip_empty, skip_managed)

    def get_next_group(self, skip_empty=False, skip_managed=False):
        return self._get_group(1, skip_empty, skip_managed)

    @expose_command()
    def unminimize_all(self):
        """Unminimise all windows in this group"""
        for win in self.windows:
            win.minimized = False
        self.layout_all()

    @expose_command()
    def next_window(self):
        """
        Focus the next window in group.

        Method cycles _all_ windows in group regardless if tiled in current
        layout or floating. Cycling of tiled and floating windows is not mixed.
        The cycling order depends on the current Layout.
        """
        if not self.windows:
            return
        if self.current_window.floating:
            nxt = (
                self.floating_layout.focus_next(self.current_window)
                or self.layout.focus_first()
                or self.floating_layout.focus_first(group=self)
            )
        else:
            nxt = (
                self.layout.focus_next(self.current_window)
                or self.floating_layout.focus_first(group=self)
                or self.layout.focus_first()
            )
        self.focus(nxt, True)

    @expose_command()
    def prev_window(self):
        """
        Focus the previous window in group.

        Method cycles _all_ windows in group regardless if tiled in current
        layout or floating. Cycling of tiled and floating windows is not mixed.
        The cycling order depends on the current Layout.
        """
        if not self.windows:
            return
        if self.current_window.floating:
            nxt = (
                self.floating_layout.focus_previous(self.current_window)
                or self.layout.focus_last()
                or self.floating_layout.focus_last(group=self)
            )
        else:
            nxt = (
                self.layout.focus_previous(self.current_window)
                or self.floating_layout.focus_last(group=self)
                or self.layout.focus_last()
            )
        self.focus(nxt, True)

    @expose_command()
    def focus_back(self):
        """
        Focus the window that had focus before the current one got it.

        Repeated calls to this function would basically continuously switch
        between the last two focused windows. Do nothing if less than 2
        windows ever received focus.
        """
        try:
            win = self.focus_history[-2]
        except IndexError:
            pass
        else:
            self.focus(win)

    @expose_command()
    def focus_by_name(self, name):
        """
        Focus the first window with the given name. Do nothing if the name is
        not found.
        """
        for win in self.windows:
            if win.name == name:
                self.focus(win)
                break

    @expose_command()
    def info_by_name(self, name):
        """
        Get the info for the first window with the given name without giving it
        focus. Do nothing if the name is not found.
        """
        for win in self.windows:
            if win.name == name:
                return win.info()

    @expose_command()
    def focus_by_index(self, index: int) -> None:
        """
        Change to the window at the specified index in the current group.
        """
        windows = self.windows
        if index < 0 or index > len(windows) - 1:
            return

        self.focus(windows[index])

    @expose_command()
    def swap_window_order(self, new_location: int) -> None:
        """
        Change the order of the current window within the current group.
        """
        if new_location < 0 or new_location > len(self.windows) - 1:
            return

        windows = self.windows
        current_window_index = windows.index(self.current_window)

        windows[current_window_index], windows[new_location] = (
            windows[new_location],
            windows[current_window_index],
        )

    @expose_command()
    def switch_groups(self, name):
        """Switch position of current group with name"""
        self.qtile.switch_groups(self.name, name)

    @expose_command()
    def set_label(self, label):
        """
        Set the display name of current group to be used in GroupBox widget.
        If label is None, the name of the group is used as display name.
        If label is the empty string, the group is invisible in GroupBox.
        """
        self.label = label if label is not None else self.name
        hook.fire("changegroup")

    def __repr__(self):
        return f"<group.Group ({self.name!r})>"



from __future__ import annotations

import os
import sys
import typing
import warnings
from logging import WARNING, Formatter, StreamHandler, captureWarnings, getLogger
from logging.handlers import RotatingFileHandler
from pathlib import Path

if typing.TYPE_CHECKING:
    from logging import Logger, LogRecord

logger = getLogger(__package__)


class ColorFormatter(Formatter):
    """Logging formatter adding console colors to the output."""

    black, red, green, yellow, blue, magenta, cyan, white = range(8)
    colors = {
        "WARNING": yellow,
        "INFO": green,
        "DEBUG": blue,
        "CRITICAL": yellow,
        "ERROR": red,
        "RED": red,
        "GREEN": green,
        "YELLOW": yellow,
        "BLUE": blue,
        "MAGENTA": magenta,
        "CYAN": cyan,
        "WHITE": white,
    }
    reset_seq = "\033[0m"
    color_seq = "\033[%dm"
    bold_seq = "\033[1m"

    def format(self, record: LogRecord) -> str:
        """Format the record with colors."""
        color = self.color_seq % (30 + self.colors[record.levelname])
        message = Formatter.format(self, record)
        message = (
            message.replace("$RESET", self.reset_seq)
            .replace("$BOLD", self.bold_seq)
            .replace("$COLOR", color)
        )
        for color, value in self.colors.items():
            message = (
                message.replace("$" + color, self.color_seq % (value + 30))
                .replace("$BG" + color, self.color_seq % (value + 40))
                .replace("$BG-" + color, self.color_seq % (value + 40))
            )
        return message + self.reset_seq


def get_default_log() -> Path:
    data_directory = os.path.expandvars("$XDG_DATA_HOME")
    if data_directory == "$XDG_DATA_HOME":
        # if variable wasn't set
        data_directory = os.path.expanduser("~/.local/share")

    qtile_directory = Path(data_directory) / "qtile"
    return qtile_directory / "qtile.log"


def init_log(
    log_level: int = WARNING,
    log_path: Path | None = None,
    log_size: int = 100_000,
    log_numbackups: int = 1,
    logger: Logger = logger,
) -> None:
    for handler in logger.handlers:
        logger.removeHandler(handler)

    if log_path is None or os.getenv("QTILE_XEPHYR"):
        # During tests or interactive xephyr development, log to stdout.
        handler = StreamHandler(sys.stdout)
        formatter: Formatter = ColorFormatter(
            "$RESET$COLOR%(asctime)s $BOLD$COLOR%(name)s "
            "%(filename)s:%(funcName)s():L%(lineno)d $RESET %(message)s"
        )

    else:
        # Otherwise during normal usage, log to file.
        try:
            log_path.parent.mkdir(parents=True, exist_ok=True)
        except PermissionError as e:
            # ok, only one place to write: stderr
            print(f"couldn't mkdir {log_path.parent}: {e}", file=sys.stderr)
            os._exit(1)

        handler = RotatingFileHandler(
            log_path,
            maxBytes=log_size,
            backupCount=log_numbackups,
        )
        formatter = Formatter(
            "%(asctime)s %(levelname)s %(name)s "
            "%(filename)s:%(funcName)s():L%(lineno)d %(message)s"
        )

    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.setLevel(log_level)
    # Capture everything from the warnings module.
    captureWarnings(True)
    warnings.simplefilter("always")
    logger.debug("Starting logging for Qtile")



from __future__ import annotations

import sys
import time
from datetime import UTC, datetime, timedelta, tzinfo

from libqtile.command.base import expose_command
from libqtile.log_utils import logger
from libqtile.widget import base

try:
    import pytz
except ImportError:
    pass

try:
    import dateutil.tz
except ImportError:
    pass


class Clock(base.InLoopPollText):
    """A simple but flexible text-based clock"""

    defaults = [
        ("format", "%H:%M", "A Python datetime format string"),
        ("update_interval", 1.0, "Update interval for the clock"),
        (
            "timezone",
            None,
            "The timezone to use for this clock, either as"
            ' string if pytz or dateutil is installed (e.g. "US/Central" or'
            " anything in /usr/share/zoneinfo), or as tzinfo (e.g."
            " datetime.timezone.utc). None means the system local timezone and is"
            " the default.",
        ),
    ]
    DELTA = timedelta(seconds=0.5)

    def __init__(self, **config):
        base.InLoopPollText.__init__(self, **config)
        self.add_defaults(Clock.defaults)
        self.timezone = self._lift_timezone(self.timezone)

        if self.timezone is None:
            logger.debug("Defaulting to the system local timezone.")

    def _lift_timezone(self, timezone):
        if isinstance(timezone, tzinfo):
            return timezone
        elif isinstance(timezone, str):
            # Empty string can be used to force use of system time
            if not timezone:
                return None

            # A string timezone needs to be converted to a tzinfo object
            if "pytz" in sys.modules:
                return pytz.timezone(timezone)
            elif "dateutil" in sys.modules:
                return dateutil.tz.gettz(timezone)
            else:
                logger.warning(
                    "Clock widget can not infer its timezone from a"
                    " string without pytz or dateutil. Install one"
                    " of these libraries, or give it a"
                    " datetime.tzinfo instance."
                )
        elif timezone is None:
            pass
        else:
            logger.warning("Invalid timezone value %s.", timezone)

        return None

    def tick(self):
        self.update(self.poll())
        return self.update_interval - time.time() % self.update_interval

    # adding .5 to get a proper seconds value because glib could
    # theoreticaly call our method too early and we could get something
    # like (x-1).999 instead of x.000
    def poll(self):
        if self.timezone:
            now = datetime.now(UTC).astimezone(self.timezone)
        else:
            now = datetime.now(UTC).astimezone()
        return (now + self.DELTA).strftime(self.format)

    @expose_command
    def update_timezone(self, timezone: str | tzinfo | None = None):
        """
        Force the clock to update timezone information.

        If the method is called with no arguments then the widget will reload
        the timzeone set on the computer (e.g. via ``timedatectl set-timezone ..``).
        This will have no effect if you have previously set a ``timezone`` value.

        Alternatively, you can pass a timezone string (e.g. ``"Europe/Lisbon"``) to change
        the specified timezone. Setting this to an empty string will cause the clock
        to rely on the system timezone.
        """
        self.timezone = self._lift_timezone(timezone)

        # Force python to update timezone info (e.g. if system time has changed)
        time.tzset()
        self.update(self.poll())

    @expose_command
    def use_system_timezone(self):
        """Force clock to use system timezone."""
        self.update_timezone("")



import enum
from typing import Any

from libqtile import configurable

try:
    from libqtile.backend.wayland._ffi import ffi, lib

except ModuleNotFoundError:
    print("Warning: Wayland backend not built. Backend will not run.")

    from libqtile.backend.wayland.ffi_stub import ffi, lib


class InputConfig(configurable.Configurable):
    """
    This is used to configure input devices. An instance of this class represents one
    set of settings that can be applied to an input device.

    To use this, define a dictionary called ``wl_input_rules`` in your config. The keys
    are used to match input devices, and the values are instances of this class with the
    desired settings. For example:

    .. code-block:: python

        from libqtile.backend.wayland import InputConfig

        wl_input_rules = {
            "1267:12377:ELAN1300:00 04F3:3059 Touchpad": InputConfig(left_handed=True),
            "*": InputConfig(left_handed=True, pointer_accel=True),
            "type:keyboard": InputConfig(kb_options="ctrl:nocaps,compose:ralt"),
        }

    When a input device is being configured, the most specific matching key in the
    dictionary is found and the corresponding settings are used to configure the device.
    Unique identifiers are chosen first, then ``"type:X"``, then ``"*"``.

    The command ``qtile cmd-obj -o core -f get_inputs`` can be used to get information
    about connected devices, including their identifiers.

    Options default to ``None``, leave a device's default settings intact. For
    information on what each option does, see the documenation for libinput:
    https://wayland.freedesktop.org/libinput/doc/latest/configuration.html. Note that
    devices often only support a subset of settings.

    This tries to mirror how Sway configures libinput devices. For more information
    check out sway-input(5): https://man.archlinux.org/man/sway-input.5#LIBINPUT_CONFIGURATION

    Keyboards, managed by `xkbcommon <https://github.com/xkbcommon/libxkbcommon>`_, are
    configured with the options prefixed by ``kb_``. X11's helpful `XKB guide
    <https://www.x.org/releases/X11R7.5/doc/input/XKB-Config.html>`_ may be useful for
    figuring out the syntax for some of these settings.
    """

    defaults = [
        ("accel_profile", None, "``'adaptive'`` or ``'flat'``"),
        ("click_method", None, "``'none'``, ``'button_areas'`` or ``'clickfinger'``"),
        ("drag", None, "``True`` or ``False``"),
        ("drag_lock", None, "``True`` or ``False``"),
        ("dwt", None, "True or False"),
        ("left_handed", None, "``True`` or ``False``"),
        ("middle_emulation", None, "``True`` or ``False``"),
        ("natural_scroll", None, "``True`` or ``False``"),
        ("pointer_accel", None, "A ``float`` between -1 and 1."),
        ("scroll_button", None, "``'disable'``, 'Button[1-3,8,9]' or a keycode"),
        (
            "scroll_method",
            None,
            "``'none'``, ``'two_finger'``, ``'edge'``, or ``'on_button_down'``",
        ),
        ("tap", None, "``True`` or ``False``"),
        ("tap_button_map", None, "``'lrm'`` or ``'lmr'``"),
        ("kb_layout", None, "Keyboard layout i.e. ``XKB_DEFAULT_LAYOUT``"),
        ("kb_options", None, "Keyboard options i.e. ``XKB_DEFAULT_OPTIONS``"),
        ("kb_variant", None, "Keyboard variant i.e. ``XKB_DEFAULT_VARIANT``"),
        ("kb_repeat_rate", 25, "Keyboard key repeats made per second"),
        ("kb_repeat_delay", 600, "Keyboard delay in milliseconds before repeating"),
    ]

    def __init__(self, **config: Any) -> None:
        configurable.Configurable.__init__(self, **config)
        self.add_defaults(InputConfig.defaults)


class InputDeviceType(enum.IntEnum):
    KEYBOARD = lib.WLR_INPUT_DEVICE_KEYBOARD
    POINTER = lib.WLR_INPUT_DEVICE_POINTER
    TOUCH = lib.WLR_INPUT_DEVICE_TOUCH
    TABLET = lib.WLR_INPUT_DEVICE_TABLET
    TABLET_PAD = lib.WLR_INPUT_DEVICE_TABLET_PAD
    SWITCH = lib.WLR_INPUT_DEVICE_SWITCH


class AccelProfile(enum.IntEnum):
    adaptive = lib.LIBINPUT_CONFIG_ACCEL_PROFILE_ADAPTIVE
    flat = lib.LIBINPUT_CONFIG_ACCEL_PROFILE_FLAT


class ClickMethod(enum.IntEnum):
    none = lib.LIBINPUT_CONFIG_CLICK_METHOD_NONE
    button_areas = lib.LIBINPUT_CONFIG_CLICK_METHOD_BUTTON_AREAS
    clickfinger = lib.LIBINPUT_CONFIG_CLICK_METHOD_CLICKFINGER


class TapMap(enum.IntEnum):
    lrm = lib.LIBINPUT_CONFIG_TAP_MAP_LRM
    lmr = lib.LIBINPUT_CONFIG_TAP_MAP_LMR


class ScrollMethod(enum.IntEnum):
    none = lib.LIBINPUT_CONFIG_SCROLL_NO_SCROLL
    two_finger = lib.LIBINPUT_CONFIG_SCROLL_2FG
    edge = lib.LIBINPUT_CONFIG_SCROLL_EDGE
    on_button_down = lib.LIBINPUT_CONFIG_SCROLL_ON_BUTTON_DOWN


def input_get_info(
    input_device: ffi.CData, name: str, type: int, vendor: int, product: int
) -> tuple[str, str]:
    if name == " " or not name.isprintable():
        name = "_"
    type_key = "type:" + InputDeviceType(type).name.lower()
    identifier = f"{vendor:d}:{product:d}:{name!s}"

    if type_key == "type:pointer" and lib is not None:
        # This checks whether the pointer is a touchpad, so that we can target those
        # specifically.
        if lib.qw_input_device_is_touchpad(input_device):
            type_key = "type:touchpad"

    return type_key, identifier


def configure_input_devices(server: ffi.CData, configs: dict[str, Any]) -> None:
    @ffi.callback(
        "void(struct qw_input_device *input_device, char *name, int type, int vendor, int product)"
    )
    def input_device_cb(
        input_device: ffi.CData, name: ffi.CData, type: int, vendor: int, product: int
    ) -> None:
        # Get the device type and identifier for this input device. These can be used be
        # used to assign ``InputConfig`` options to devices or types of devices.
        name_dec = ffi.string(name).decode()
        type_key, identifier = input_get_info(input_device, name_dec, type, vendor, product)
        if identifier in configs:
            conf = configs[identifier]
        elif type_key in configs:
            conf = configs[type_key]
        elif "*" in configs:
            conf = configs["*"]
        else:
            conf = None

        if conf is not None:
            if type == InputDeviceType.POINTER:
                device = lib.qw_input_device_get_libinput_handle(input_device)
                if device == ffi.NULL:
                    return

                if conf.accel_profile is not None:
                    lib.qw_input_device_config_accel_set_profile(
                        device, AccelProfile[conf.accel_profile].value
                    )

                if conf.pointer_accel is not None:
                    lib.qw_input_device_config_accel_set_speed(device, conf.pointer_accel)

                if conf.click_method is not None:
                    lib.qw_input_device_config_click_set_method(
                        device, ClickMethod[conf.click_method].value
                    )

                if conf.drag is not None:
                    lib.qw_input_device_config_tap_set_drag_enabled(device, int(conf.drag))

                if conf.drag_lock is not None:
                    lib.qw_input_device_config_tap_set_drag_lock_enabled(
                        device, int(conf.drag_lock)
                    )

                if conf.tap is not None:
                    lib.qw_input_device_config_tap_set_enabled(device, int(conf.tap))

                if conf.tap_button_map is not None:
                    lib.qw_input_device_config_tap_set_button_map(
                        device, TapMap[conf.tap_button_map].value
                    )

                if conf.natural_scroll is not None:
                    lib.qw_input_device_config_scroll_set_natural_scroll_enabled(
                        device, int(conf.natural_scroll)
                    )

                if conf.scroll_method is not None:
                    lib.qw_input_device_config_scroll_set_method(
                        device, ScrollMethod[conf.scroll_method].value
                    )

                if conf.scroll_button is not None:
                    if isinstance(conf.scroll_button, str):
                        if conf.scroll_button == "disable":
                            button = 0
                        else:
                            button = lib.qw_util_get_button_code(int(conf.scroll_button[-1]) - 1)
                    else:
                        button = conf.scroll_button
                    lib.qw_input_device_config_scroll_set_button(device, button)

                if conf.dwt is not None:
                    lib.qw_input_device_config_dwt_set_enabled(device, int(conf.dwt))

                if conf.left_handed is not None:
                    lib.qw_input_device_config_left_handed_set(device, int(conf.left_handed))

                if conf.middle_emulation is not None:
                    lib.qw_input_device_config_middle_emulation_set_enabled(
                        device, int(conf.middle_emulation)
                    )

            elif type == InputDeviceType.KEYBOARD:
                keyboard = lib.qw_input_device_get_keyboard(input_device)
                if keyboard == ffi.NULL:
                    return

                lib.qw_keyboard_set_repeat_info(
                    keyboard, conf.kb_repeat_rate, conf.kb_repeat_delay
                )
                lib.qw_keyboard_set_keymap(
                    keyboard,
                    ffi.new("char[]", (conf.kb_layout or "").encode()),
                    ffi.new("char[]", (conf.kb_options or "").encode()),
                    ffi.new("char[]", (conf.kb_variant or "").encode()),
                )

    lib.qw_server_loop_input_devices(server, input_device_cb)



from __future__ import annotations

import difflib
from textwrap import dedent
from typing import TYPE_CHECKING

import libcst
import libcst.matchers as m

from libqtile.scripts.migrations import MIGRATIONS

if TYPE_CHECKING:
    from collections.abc import Collection
    from typing import ClassVar

    from libcst.metadata.base_provider import ProviderT


# By default, libcst adds spaces around "=" so we should remove them
EQUALS_NO_SPACE = libcst.AssignEqual(
    whitespace_before=libcst.SimpleWhitespace(
        value="",
    ),
    whitespace_after=libcst.SimpleWhitespace(
        value="",
    ),
)

RED = "\033[38;2;255;0;0m"
GREEN = "\033[38;2;0;255;0m"
END = "\033[38;2;255;255;255m"


def add_migration(migration):
    MIGRATIONS.append(migration)


def add_colours(lines):
    coloured = []
    for line in lines:
        if line.startswith("+"):
            coloured.append(f"{GREEN}{line}{END}")
        elif line.startswith("-"):
            coloured.append(f"{RED}{line}{END}")
        else:
            coloured.append(line)

    return coloured


class LintLine:
    def __init__(self, pos, message: str = "", migrator_id: str = "") -> None:
        self.line = pos.line
        self.col = pos.column
        self.message = message
        self.id = migrator_id
        self._pos = (pos.line, pos.column)

    def set_id(self, migration_id: str = "") -> LintLine:
        self.id = migration_id
        return self

    def __lt__(self, other: LintLine) -> bool:
        if not isinstance(other, LintLine):
            raise ValueError
        return self._pos < other._pos

    def __str__(self) -> str:
        return "[Ln {line}, Col {col}]: {message} ({id})".format(**self.__dict__)


class MigrationHelper:
    """
    Migrations may need to access details about the position of the match (e.g. for linting) and
    the parent node.

    These are accessible via metadata providers so this class sets the relevant dependencies
    and provides some helper functions.
    """

    METADATA_DEPENDENCIES: ClassVar[Collection[ProviderT]] = (
        libcst.metadata.PositionProvider,
        libcst.metadata.ParentNodeProvider,
    )

    def __init__(self) -> None:
        self._lint_lines: list[LintLine] = []

    def lint(self, node, message) -> None:
        pos = self.get_metadata(libcst.metadata.PositionProvider, node).start  # type: ignore[attr-defined]
        self._lint_lines.append(LintLine(pos, message))

    def get_parent(self, node) -> libcst.CSTNode:
        return self.get_metadata(libcst.metadata.ParentNodeProvider, node)  # type: ignore[attr-defined]


class MigrationTransformer(MigrationHelper, m.MatcherDecoratableTransformer):
    def __init__(self, *args, **kwargs):
        m.MatcherDecoratableTransformer.__init__(self, *args, **kwargs)
        MigrationHelper.__init__(self)


class MigrationVisitor(MigrationHelper, m.MatcherDecoratableVisitor):
    def __init__(self, *args, **kwargs):
        m.MatcherDecoratableVisitor.__init__(self, *args, **kwargs)
        MigrationHelper.__init__(self)


class RenamerTransformer(MigrationTransformer):
    """
    Very basic transformer. Will replace every instance of 'old_name'
    with 'new_name'.

    To use: subclass RenameTransformer and set the class attribute
    'from_to' with ('old_name', 'new_name')
    """

    from_to = ("old_name", "new_name")

    def leave_Name(self, original_node, updated_node):  # noqa: N802
        if original_node.value == self.from_to[0]:
            self.lint(original_node, "'{}' should be replaced with '{}'.".format(*self.from_to))
            return updated_node.with_changes(value=self.from_to[1])
        return original_node


class Change:
    def __init__(self, input_code, output_code):
        self.input = input_code
        self.output = output_code
        self.check = False


class NoChange(Change):
    def __init__(self, input_code):
        Change.__init__(self, input_code, input_code)


class Check(Change):
    def __init__(self, input_code, output_code):
        Change.__init__(self, input_code, output_code)
        self.check = True


class _QtileMigrator:
    ID: str = "UniqueMigrationName"

    SUMMARY: str = "Brief summary of migration"

    HELP: str = """
    More detailed description of the migration including purpose and example changes.
    """

    AFTER_VERSION: str = ""

    TESTS: list[Change] = []

    # Set the name of the visitor/transformer here
    visitor: type[MigrationVisitor] | type[MigrationTransformer] | None = None

    def __init__(self) -> None:
        self._lint_lines: list[LintLine] = []
        self.original: libcst.metadata.MetadataWrapper | None = None
        self.updated: libcst.Module | None = None

    def update_lint(self, visitor) -> None:
        self._lint_lines += [line.set_id(self.ID) for line in visitor._lint_lines]

    def show_lint(self) -> list[LintLine]:
        """
        Return a list of LintLine objects storing details of deprecated code.
        """
        return self._lint_lines

    @classmethod
    def show_help(cls) -> str:
        return dedent(cls.HELP)

    @classmethod
    def show_summary(cls) -> str:
        return cls.SUMMARY

    @classmethod
    def show_id(cls) -> str:
        return cls.ID

    @classmethod
    def get_version(cls) -> tuple[int, ...]:
        return tuple(int(x) for x in cls.AFTER_VERSION.split("."))

    def show_diff(self, no_colour=False) -> str:
        """
        Return a string showing the generated diff.
        """
        if self.original is None or self.updated is None:
            return ""

        original = self.original.module
        updated = self.updated

        diff = difflib.unified_diff(
            original.code.splitlines(), updated.code.splitlines(), "original", "modified"
        )

        if not no_colour:
            diff = add_colours(list(diff))

        try:
            a, b, *lines = diff
            joined = "\n".join(lines)
            return f"{a}{b}{joined}"
        except ValueError:
            # No difference so we stop here.
            return ""

    def migrate(self, original) -> None:
        original, updated = self.run(original)
        self.original = original
        self.updated = updated

    def run(
        self, original
    ) -> tuple[libcst.Module | libcst.metadata.MetadataWrapper, libcst.Module]:
        """
        Run the migration set in 'visitor'.

        Method can be overriden for more complex scenarios (e.g. multiple transformers).
        """
        if self.visitor is None:
            raise AttributeError("Migration has not defined a 'visitor' attribute.")

        transformer = self.visitor()
        updated = original.visit(transformer)
        self.update_lint(transformer)

        return original, updated



from libqtile import DynamicLibraries, find_library
from libqtile.backend.x11.xcursors_ffi import xcursors_ffi as ffi
from libqtile.log_utils import logger


# Stolen from samurai-x
# (Don't know where to put it, so I'll put it here)
# XCB cursors doesn't want to be themed, libxcursor
# would be better choice I think
# and we (indirectly) depend on it anyway...
class Cursors(dict):
    def __init__(self, conn):
        self.conn = conn
        cursors = (
            (b"X_cursor", 0),
            (b"arrow", 2),
            (b"based_arrow_down", 4),
            (b"based_arrow_up", 6),
            (b"boat", 8),
            (b"bogosity", 10),
            (b"bottom_left_corner", 12),
            (b"bottom_right_corner", 14),
            (b"bottom_side", 16),
            (b"bottom_tee", 18),
            (b"box_spiral", 20),
            (b"center_ptr", 22),
            (b"circle", 24),
            (b"clock", 26),
            (b"coffee_mug", 28),
            (b"cross", 30),
            (b"cross_reverse", 32),
            (b"crosshair", 34),
            (b"diamond_cross", 36),
            (b"dot", 38),
            (b"dotbox", 40),
            (b"double_arrow", 42),
            (b"draft_large", 44),
            (b"draft_small", 46),
            (b"draped_box", 48),
            (b"exchange", 50),
            (b"fleur", 52),
            (b"gobbler", 54),
            (b"gumby", 56),
            (b"hand1", 58),
            (b"hand2", 60),
            (b"heart", 62),
            (b"icon", 64),
            (b"iron_cross", 66),
            (b"left_ptr", 68),
            (b"left_side", 70),
            (b"left_tee", 72),
            (b"leftbutton", 74),
            (b"ll_angle", 76),
            (b"lr_angle", 78),
            (b"man", 80),
            (b"middlebutton", 82),
            (b"mouse", 84),
            (b"pencil", 86),
            (b"pirate", 88),
            (b"plus", 90),
            (b"question_arrow", 92),
            (b"right_ptr", 94),
            (b"right_side", 96),
            (b"right_tee", 98),
            (b"rightbutton", 100),
            (b"rtl_logo", 102),
            (b"sailboat", 104),
            (b"sb_down_arrow", 106),
            (b"sb_h_double_arrow", 108),
            (b"sb_left_arrow", 110),
            (b"sb_right_arrow", 112),
            (b"sb_up_arrow", 114),
            (b"sb_v_double_arrow", 116),
            (b"shuttle", 118),
            (b"sizing", 120),
            (b"spider", 122),
            (b"spraycan", 124),
            (b"star", 126),
            (b"target", 128),
            (b"tcross", 130),
            (b"top_left_arrow", 132),
            (b"top_left_corner", 134),
            (b"top_right_corner", 136),
            (b"top_side", 138),
            (b"top_tee", 140),
            (b"trek", 142),
            (b"ul_angle", 144),
            (b"umbrella", 146),
            (b"ur_angle", 148),
            (b"watch", 150),
            (b"xterm", 152),
        )

        self.xcursor = self._setup_xcursor_binding()

        for name, cursor_font in cursors:
            self._new(name, cursor_font)

        if self.xcursor:
            self.xcursor.xcb_cursor_context_free(self._cursor_ctx[0])

    def finalize(self):
        self._cursor_ctx = None

    def _setup_xcursor_binding(self):
        lib = find_library(DynamicLibraries.XCBCURSOR)
        if lib is None:
            logger.info("xcb-cursor not found, fallback to font pointer")
            return False

        xcursor = ffi.dlopen(lib)

        conn = self.conn.conn
        screen_pointer = conn.get_screen_pointers()[0]
        self._cursor_ctx = ffi.new("xcb_cursor_context_t **")
        xcursor.xcb_cursor_context_new(conn._conn, screen_pointer, self._cursor_ctx)

        return xcursor

    def get_xcursor(self, name):
        """
        Get the cursor using xcb-util-cursor, so we support themed cursors
        """
        cursor = self.xcursor.xcb_cursor_load_cursor(self._cursor_ctx[0], name)
        return cursor

    def get_font_cursor(self, name, cursor_font):
        """
        Get the cursor from the font, used as a fallback if xcb-util-cursor
        is not installed
        """
        fid = self.conn.conn.generate_id()
        self.conn.conn.core.OpenFont(fid, len("cursor"), "cursor")
        cursor = self.conn.conn.generate_id()
        self.conn.conn.core.CreateGlyphCursor(
            cursor, fid, fid, cursor_font, cursor_font + 1, 0, 0, 0, 65535, 65535, 65535
        )
        return cursor

    def _new(self, name, cursor_font):
        if self.xcursor:
            cursor = self.get_xcursor(name)
        else:
            cursor = self.get_font_cursor(name, cursor_font)
        self[name.decode()] = cursor
